# cython: infer_types=True, binding=True
import importlib
import sys
from collections import defaultdict
from typing import Callable, Optional

from thinc.api import Config, Model

from ._parser_internals.transition_system import TransitionSystem

from ._parser_internals.arc_eager cimport ArcEager
from .transition_parser cimport Parser

from ..language import Language
from ..scorer import Scorer
from ..training import remove_bilu_prefix
from ..util import registry
from ._parser_internals import nonproj
from ._parser_internals.nonproj import DELIMITER
from .functions import merge_subtokens

default_model_config = """
[model]
@architectures = "spacy.TransitionBasedParser.v2"
state_type = "parser"
extra_state_tokens = false
hidden_width = 64
maxout_pieces = 2
use_upper = true

[model.tok2vec]
@architectures = "spacy.HashEmbedCNN.v2"
pretrained_vectors = null
width = 96
depth = 4
embed_size = 2000
window_size = 1
maxout_pieces = 3
subword_features = true
"""
DEFAULT_PARSER_MODEL = Config().from_str(default_model_config)["model"]


def parser_score(examples, **kwargs):
    """Score a batch of examples.

    examples (Iterable[Example]): The examples to score.
    RETURNS (Dict[str, Any]): The scores, produced by Scorer.score_spans
        and Scorer.score_deps.

    DOCS: https://spacy.io/api/dependencyparser#score
    """
    def has_sents(doc):
        return doc.has_annotation("SENT_START")

    def dep_getter(token, attr):
        dep = getattr(token, attr)
        dep = token.vocab.strings.as_string(dep).lower()
        return dep
    results = {}
    results.update(Scorer.score_spans(examples, "sents", has_annotation=has_sents, **kwargs))
    kwargs.setdefault("getter", dep_getter)
    kwargs.setdefault("ignore_labels", ("p", "punct"))
    results.update(Scorer.score_deps(examples, "dep", **kwargs))
    del results["sents_per_type"]
    return results


def make_parser_scorer():
    return parser_score


cdef class DependencyParser(Parser):
    """Pipeline component for dependency parsing.

    DOCS: https://spacy.io/api/dependencyparser
    """
    TransitionSystem = ArcEager

    def __init__(
        self,
        vocab,
        model,
        name="parser",
        moves=None,
        *,
        update_with_oracle_cut_size=100,
        min_action_freq=30,
        learn_tokens=False,
        beam_width=1,
        beam_density=0.0,
        beam_update_prob=0.0,
        multitasks=tuple(),
        incorrect_spans_key=None,
        scorer=parser_score,
    ):
        """Create a DependencyParser.
        """
        super().__init__(
            vocab,
            model,
            name,
            moves,
            update_with_oracle_cut_size=update_with_oracle_cut_size,
            min_action_freq=min_action_freq,
            learn_tokens=learn_tokens,
            beam_width=beam_width,
            beam_density=beam_density,
            beam_update_prob=beam_update_prob,
            multitasks=multitasks,
            incorrect_spans_key=incorrect_spans_key,
            scorer=scorer,
        )

    @property
    def postprocesses(self):
        output = [nonproj.deprojectivize]
        if self.cfg.get("learn_tokens") is True:
            output.append(merge_subtokens)
        return tuple(output)

    def add_multitask_objective(self, mt_component):
        self._multitasks.append(mt_component)

    def init_multitask_objectives(self, get_examples, nlp=None, **cfg):
        # TODO: transfer self.model.get_ref("tok2vec") to the multitask's model ?
        for labeller in self._multitasks:
            labeller.model.set_dim("nO", len(self.labels))
            if labeller.model.has_ref("output_layer"):
                labeller.model.get_ref("output_layer").set_dim("nO", len(self.labels))
            labeller.initialize(get_examples, nlp=nlp)

    @property
    def labels(self):
        labels = set()
        # Get the labels from the model by looking at the available moves
        for move in self.move_names:
            if "-" in move:
                label = remove_bilu_prefix(move)
                if DELIMITER in label:
                    label = label.split(DELIMITER)[1]
                labels.add(label)
        return tuple(sorted(labels))

    def scored_parses(self, beams):
        """Return two dictionaries with scores for each beam/doc that was processed:
        one containing (i, head) keys, and another containing (i, label) keys.
        """
        head_scores = []
        label_scores = []
        for beam in beams:
            score_head_dict = defaultdict(float)
            score_label_dict = defaultdict(float)
            for score, parses in self.moves.get_beam_parses(beam):
                for head, i, label in parses:
                    score_head_dict[(i, head)] += score
                    score_label_dict[(i, label)] += score
            head_scores.append(score_head_dict)
            label_scores.append(score_label_dict)
        return head_scores, label_scores

    def _ensure_labels_are_added(self, docs):
        # This gives the parser a chance to add labels it's missing for a batch
        # of documents. However, this isn't desirable for the dependency parser,
        # because we instead have a label frequency cut-off and back off rare
        # labels to 'dep'.
        pass


# Setup backwards compatibility hook for factories
def __getattr__(name):
    if name == "make_parser":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_parser
    elif name == "make_beam_parser":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_beam_parser
    raise AttributeError(f"module {__name__} has no attribute {name}")



from ...language import BaseDefaults, Language
from .lex_attrs import LEX_ATTRS
from .stop_words import STOP_WORDS


class KurmanjiDefaults(BaseDefaults):
    stop_words = STOP_WORDS
    lex_attr_getters = LEX_ATTRS


class Kurmanji(Language):
    lang = "kmr"
    Defaults = KurmanjiDefaults


__all__ = ["Kurmanji"]



"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.bo.examples import sentences
>>> docs = nlp.pipe(sentences)
"""

sentences = [
    "དོན་དུ་རྒྱ་མཚོ་བླ་མ་ཞེས་བྱ་ཞིང༌།",
    "ཏཱ་ལའི་ཞེས་པ་ནི་སོག་སྐད་ཡིན་པ་དེ་བོད་སྐད་དུ་རྒྱ་མཚོའི་དོན་དུ་འཇུག",
    "སོག་པོ་ཨལ་ཐན་རྒྱལ་པོས་རྒྱལ་དབང་བསོད་ནམས་རྒྱ་མཚོར་ཆེ་བསྟོད་ཀྱི་མཚན་གསོལ་བ་ཞིག་ཡིན་ཞིང༌།",
    "རྗེས་སུ་རྒྱལ་བ་དགེ་འདུན་གྲུབ་དང༌། དགེ་འདུན་རྒྱ་མཚོ་སོ་སོར་ཡང་ཏཱ་ལའི་བླ་མའི་སྐུ་ཕྲེང་དང་པོ་དང༌།",
    "གཉིས་པའི་མཚན་དེ་གསོལ་ཞིང༌།༸རྒྱལ་དབང་སྐུ་ཕྲེང་ལྔ་པས་དགའ་ལྡན་ཕོ་བྲང་གི་སྲིད་དབང་བཙུགས་པ་ནས་ཏཱ་ལའི་བླ་མ་ནི་བོད་ཀྱི་ཆོས་སྲིད་གཉིས་ཀྱི་དབུ་ཁྲིད་དུ་གྱུར་ཞིང་།",
    "ད་ལྟའི་བར་ཏཱ་ལའི་བླ་མ་སྐུ་ཕྲེང་བཅུ་བཞི་བྱོན་ཡོད།",
]



"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.ht.examples import sentences
>>> docs = nlp.pipe(sentences)
"""

sentences = [
    "Apple ap panse achte yon demaraj nan Wayòm Ini pou $1 milya dola",
    "Machin otonòm fè responsablite asirans lan ale sou men fabrikan yo",
    "San Francisco ap konsidere entèdi robo ki livre sou twotwa yo",
    "Lond se yon gwo vil nan Wayòm Ini",
    "Kote ou ye?",
    "Kilès ki prezidan Lafrans?",
    "Ki kapital Etazini?",
    "Kile Barack Obama te fèt?",
]



# cython: infer_types=True, binding=True
import importlib
import sys
from collections import defaultdict
from typing import Callable, Optional

from thinc.api import Config, Model

from ._parser_internals.transition_system import TransitionSystem

from ._parser_internals.ner cimport BiluoPushDown
from .transition_parser cimport Parser

from ..language import Language
from ..scorer import get_ner_prf
from ..training import remove_bilu_prefix
from ..util import registry

default_model_config = """
[model]
@architectures = "spacy.TransitionBasedParser.v2"
state_type = "ner"
extra_state_tokens = false
hidden_width = 64
maxout_pieces = 2
use_upper = true

[model.tok2vec]
@architectures = "spacy.HashEmbedCNN.v2"
pretrained_vectors = null
width = 96
depth = 4
embed_size = 2000
window_size = 1
maxout_pieces = 3
subword_features = true
"""
DEFAULT_NER_MODEL = Config().from_str(default_model_config)["model"]


def ner_score(examples, **kwargs):
    return get_ner_prf(examples, **kwargs)


def make_ner_scorer():
    return ner_score


cdef class EntityRecognizer(Parser):
    """Pipeline component for named entity recognition.

    DOCS: https://spacy.io/api/entityrecognizer
    """
    TransitionSystem = BiluoPushDown

    def __init__(
        self,
        vocab,
        model,
        name="ner",
        moves=None,
        *,
        update_with_oracle_cut_size=100,
        beam_width=1,
        beam_density=0.0,
        beam_update_prob=0.0,
        multitasks=tuple(),
        incorrect_spans_key=None,
        scorer=ner_score,
    ):
        """Create an EntityRecognizer.
        """
        super().__init__(
            vocab,
            model,
            name,
            moves,
            update_with_oracle_cut_size=update_with_oracle_cut_size,
            min_action_freq=1,   # not relevant for NER
            learn_tokens=False,  # not relevant for NER
            beam_width=beam_width,
            beam_density=beam_density,
            beam_update_prob=beam_update_prob,
            multitasks=multitasks,
            incorrect_spans_key=incorrect_spans_key,
            scorer=scorer,
        )

    def add_multitask_objective(self, mt_component):
        """Register another component as a multi-task objective. Experimental."""
        self._multitasks.append(mt_component)

    def init_multitask_objectives(self, get_examples, nlp=None, **cfg):
        """Setup multi-task objective components. Experimental and internal."""
        # TODO: transfer self.model.get_ref("tok2vec") to the multitask's model ?
        for labeller in self._multitasks:
            labeller.model.set_dim("nO", len(self.labels))
            if labeller.model.has_ref("output_layer"):
                labeller.model.get_ref("output_layer").set_dim("nO", len(self.labels))
            labeller.initialize(get_examples, nlp=nlp)

    @property
    def labels(self):
        # Get the labels from the model by looking at the available moves, e.g.
        # B-PERSON, I-PERSON, L-PERSON, U-PERSON
        labels = set(remove_bilu_prefix(move) for move in self.move_names
                     if move[0] in ("B", "I", "L", "U"))
        return tuple(sorted(labels))

    def scored_ents(self, beams):
        """Return a dictionary of (start, end, label) tuples with corresponding scores
        for each beam/doc that was processed.
        """
        entity_scores = []
        for beam in beams:
            score_dict = defaultdict(float)
            for score, ents in self.moves.get_beam_parses(beam):
                for start, end, label in ents:
                    score_dict[(start, end, label)] += score
            entity_scores.append(score_dict)
        return entity_scores


# Setup backwards compatibility hook for factories
def __getattr__(name):
    if name == "make_ner":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_ner
    elif name == "make_beam_ner":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_beam_ner
    raise AttributeError(f"module {__name__} has no attribute {name}")



from spacy.symbols import NORM, ORTH


def make_variants(base, first_norm, second_orth, second_norm):
    return {
        base: [
            {ORTH: base.split("'")[0] + "'", NORM: first_norm},
            {ORTH: second_orth, NORM: second_norm},
        ],
        base.capitalize(): [
            {
                ORTH: base.split("'")[0].capitalize() + "'",
                NORM: first_norm.capitalize(),
            },
            {ORTH: second_orth, NORM: second_norm},
        ],
    }


TOKENIZER_EXCEPTIONS = {"Dr.": [{ORTH: "Dr."}]}

# Apostrophe forms
TOKENIZER_EXCEPTIONS.update(make_variants("m'ap", "mwen", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("n'ap", "nou", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("l'ap", "li", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("y'ap", "yo", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("m'te", "mwen", "te", "te"))
TOKENIZER_EXCEPTIONS.update(make_variants("m'pral", "mwen", "pral", "pral"))
TOKENIZER_EXCEPTIONS.update(make_variants("w'ap", "ou", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("k'ap", "ki", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("p'ap", "pa", "ap", "ap"))
TOKENIZER_EXCEPTIONS.update(make_variants("t'ap", "te", "ap", "ap"))

# Non-apostrophe contractions (with capitalized variants)
TOKENIZER_EXCEPTIONS.update(
    {
        "map": [
            {ORTH: "m", NORM: "mwen"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Map": [
            {ORTH: "M", NORM: "Mwen"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "lem": [
            {ORTH: "le", NORM: "le"},
            {ORTH: "m", NORM: "mwen"},
        ],
        "Lem": [
            {ORTH: "Le", NORM: "Le"},
            {ORTH: "m", NORM: "mwen"},
        ],
        "lew": [
            {ORTH: "le", NORM: "le"},
            {ORTH: "w", NORM: "ou"},
        ],
        "Lew": [
            {ORTH: "Le", NORM: "Le"},
            {ORTH: "w", NORM: "ou"},
        ],
        "nap": [
            {ORTH: "n", NORM: "nou"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Nap": [
            {ORTH: "N", NORM: "Nou"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "lap": [
            {ORTH: "l", NORM: "li"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Lap": [
            {ORTH: "L", NORM: "Li"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "yap": [
            {ORTH: "y", NORM: "yo"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Yap": [
            {ORTH: "Y", NORM: "Yo"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "mte": [
            {ORTH: "m", NORM: "mwen"},
            {ORTH: "te", NORM: "te"},
        ],
        "Mte": [
            {ORTH: "M", NORM: "Mwen"},
            {ORTH: "te", NORM: "te"},
        ],
        "mpral": [
            {ORTH: "m", NORM: "mwen"},
            {ORTH: "pral", NORM: "pral"},
        ],
        "Mpral": [
            {ORTH: "M", NORM: "Mwen"},
            {ORTH: "pral", NORM: "pral"},
        ],
        "wap": [
            {ORTH: "w", NORM: "ou"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Wap": [
            {ORTH: "W", NORM: "Ou"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "kap": [
            {ORTH: "k", NORM: "ki"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Kap": [
            {ORTH: "K", NORM: "Ki"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "tap": [
            {ORTH: "t", NORM: "te"},
            {ORTH: "ap", NORM: "ap"},
        ],
        "Tap": [
            {ORTH: "T", NORM: "Te"},
            {ORTH: "ap", NORM: "ap"},
        ],
    }
)



from ...attrs import LIKE_NUM

_num_words = [
    "sifir",
    "yek",
    "du",
    "sê",
    "çar",
    "pênc",
    "şeş",
    "heft",
    "heşt",
    "neh",
    "deh",
    "yazde",
    "dazde",
    "sêzde",
    "çarde",
    "pazde",
    "şazde",
    "hevde",
    "hejde",
    "nozde",
    "bîst",
    "sî",
    "çil",
    "pêncî",
    "şêst",
    "heftê",
    "heştê",
    "nod",
    "sed",
    "hezar",
    "milyon",
    "milyar",
]

_ordinal_words = [
    "yekem",
    "yekemîn",
    "duyem",
    "duyemîn",
    "sêyem",
    "sêyemîn",
    "çarem",
    "çaremîn",
    "pêncem",
    "pêncemîn",
    "şeşem",
    "şeşemîn",
    "heftem",
    "heftemîn",
    "heştem",
    "heştemîn",
    "nehem",
    "nehemîn",
    "dehem",
    "dehemîn",
    "yazdehem",
    "yazdehemîn",
    "dazdehem",
    "dazdehemîn",
    "sêzdehem",
    "sêzdehemîn",
    "çardehem",
    "çardehemîn",
    "pazdehem",
    "pazdehemîn",
    "şanzdehem",
    "şanzdehemîn",
    "hevdehem",
    "hevdehemîn",
    "hejdehem",
    "hejdehemîn",
    "nozdehem",
    "nozdehemîn",
    "bîstem",
    "bîstemîn",
    "sîyem",
    "sîyemîn",
    "çilem",
    "çilemîn",
    "pêncîyem",
    "pênciyemîn",
    "şêstem",
    "şêstemîn",
    "heftêyem",
    "heftêyemîn",
    "heştêyem",
    "heştêyemîn",
    "notem",
    "notemîn",
    "sedem",
    "sedemîn",
    "hezarem",
    "hezaremîn",
    "milyonem",
    "milyonemîn",
    "milyarem",
    "milyaremîn",
]


def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    text_lower = text.lower()
    if text_lower in _num_words:
        return True

    # Check ordinal number
    if text_lower in _ordinal_words:
        return True

    if is_digit(text_lower):
        return True

    return False


def is_digit(text):
    endings = ("em", "yem", "emîn", "yemîn")
    for ending in endings:
        to = len(ending)
        if text.endswith(ending) and text[:-to].isdigit():
            return True

    return False


LEX_ATTRS = {LIKE_NUM: like_num}



from ..char_classes import (
    ALPHA,
    ALPHA_LOWER,
    ALPHA_UPPER,
    CONCAT_QUOTES,
    HYPHENS,
    LIST_ELLIPSES,
    LIST_ICONS,
    LIST_PUNCT,
    LIST_QUOTES,
    merge_chars,
)

ELISION = "'’".replace(" ", "")

_prefixes_elision = "m n l y t k w"
_prefixes_elision += " " + _prefixes_elision.upper()

TOKENIZER_PREFIXES = (
    LIST_PUNCT
    + LIST_QUOTES
    + [
        r"(?:({pe})[{el}])(?=[{a}])".format(
            a=ALPHA, el=ELISION, pe=merge_chars(_prefixes_elision)
        )
    ]
)

TOKENIZER_SUFFIXES = (
    LIST_PUNCT
    + LIST_QUOTES
    + LIST_ELLIPSES
    + [
        r"(?<=[0-9])%",  # numbers like 10%
        r"(?<=[0-9])(?:{h})".format(h=HYPHENS),  # hyphens after numbers
        r"(?<=[{a}])['’]".format(a=ALPHA),  # apostrophes after letters
        r"(?<=[{a}])['’][mwlnytk](?=\s|$)".format(a=ALPHA),  # contractions
        r"(?<=[{a}0-9])\)",  # right parenthesis after letter/number
        r"(?<=[{a}])\.(?=\s|$)".format(
            a=ALPHA
        ),  # period after letter if space or end of string
        r"(?<=\))[\.\?!]",  # punctuation immediately after right parenthesis
    ]
)

TOKENIZER_INFIXES = (
    LIST_ELLIPSES
    + LIST_ICONS
    + [
        r"(?<=[0-9])[+\-\*^](?=[0-9-])",
        r"(?<=[{al}{q}])\.(?=[{au}{q}])".format(
            al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES
        ),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        r"(?<=[{a}0-9])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),
        r"(?<=[{a}][{el}])(?=[{a}])".format(a=ALPHA, el=ELISION),
    ]
)



import importlib
import warnings
from typing import Any, Dict

import srsly

from .. import util
from ..errors import Warnings
from ..language import Language
from ..matcher import Matcher
from ..tokens import Doc


@Language.component(
    "merge_noun_chunks",
    requires=["token.dep", "token.tag", "token.pos"],
    retokenizes=True,
)
def merge_noun_chunks(doc: Doc) -> Doc:
    """Merge noun chunks into a single token.

    doc (Doc): The Doc object.
    RETURNS (Doc): The Doc object with merged noun chunks.

    DOCS: https://spacy.io/api/pipeline-functions#merge_noun_chunks
    """
    if not doc.has_annotation("DEP"):
        return doc
    with doc.retokenize() as retokenizer:
        for np in doc.noun_chunks:
            attrs = {"tag": np.root.tag, "dep": np.root.dep}
            retokenizer.merge(np, attrs=attrs)  # type: ignore[arg-type]
    return doc


@Language.component(
    "merge_entities",
    requires=["doc.ents", "token.ent_iob", "token.ent_type"],
    retokenizes=True,
)
def merge_entities(doc: Doc):
    """Merge entities into a single token.

    doc (Doc): The Doc object.
    RETURNS (Doc): The Doc object with merged entities.

    DOCS: https://spacy.io/api/pipeline-functions#merge_entities
    """
    with doc.retokenize() as retokenizer:
        for ent in doc.ents:
            attrs = {"tag": ent.root.tag, "dep": ent.root.dep, "ent_type": ent.label}
            retokenizer.merge(ent, attrs=attrs)  # type: ignore[arg-type]
    return doc


@Language.component("merge_subtokens", requires=["token.dep"], retokenizes=True)
def merge_subtokens(doc: Doc, label: str = "subtok") -> Doc:
    """Merge subtokens into a single token.

    doc (Doc): The Doc object.
    label (str): The subtoken dependency label.
    RETURNS (Doc): The Doc object with merged subtokens.

    DOCS: https://spacy.io/api/pipeline-functions#merge_subtokens
    """
    # TODO: make stateful component with "label" config
    merger = Matcher(doc.vocab)
    merger.add("SUBTOK", [[{"DEP": label, "op": "+"}]])
    matches = merger(doc)
    spans = util.filter_spans([doc[start : end + 1] for _, start, end in matches])  # type: ignore[misc, operator]
    with doc.retokenize() as retokenizer:
        for span in spans:
            retokenizer.merge(span)
    return doc


class TokenSplitter:
    def __init__(self, min_length: int = 0, split_length: int = 0):
        self.min_length = min_length
        self.split_length = split_length

    def __call__(self, doc: Doc) -> Doc:
        if self.min_length > 0 and self.split_length > 0:
            with doc.retokenize() as retokenizer:
                for t in doc:
                    if len(t.text) >= self.min_length:
                        orths = []
                        heads = []
                        attrs = {}  # type: ignore[var-annotated]
                        for i in range(0, len(t.text), self.split_length):
                            orths.append(t.text[i : i + self.split_length])
                            heads.append((t, i / self.split_length))
                        retokenizer.split(t, orths, heads, attrs)  # type: ignore[arg-type]
        return doc

    def _get_config(self) -> Dict[str, Any]:
        return {
            "min_length": self.min_length,
            "split_length": self.split_length,
        }

    def _set_config(self, config: Dict[str, Any] = {}) -> None:
        self.min_length = config.get("min_length", 0)
        self.split_length = config.get("split_length", 0)

    def to_bytes(self, **kwargs):
        serializers = {
            "cfg": lambda: srsly.json_dumps(self._get_config()),
        }
        return util.to_bytes(serializers, [])

    def from_bytes(self, data, **kwargs):
        deserializers = {
            "cfg": lambda b: self._set_config(srsly.json_loads(b)),
        }
        util.from_bytes(data, deserializers, [])
        return self

    def to_disk(self, path, **kwargs):
        path = util.ensure_path(path)
        serializers = {
            "cfg": lambda p: srsly.write_json(p, self._get_config()),
        }
        return util.to_disk(path, serializers, [])

    def from_disk(self, path, **kwargs):
        path = util.ensure_path(path)
        serializers = {
            "cfg": lambda p: self._set_config(srsly.read_json(p)),
        }
        util.from_disk(path, serializers, [])


class DocCleaner:
    def __init__(self, attrs: Dict[str, Any], *, silent: bool = True):
        self.cfg: Dict[str, Any] = {"attrs": dict(attrs), "silent": silent}

    def __call__(self, doc: Doc) -> Doc:
        attrs: dict = self.cfg["attrs"]
        silent: bool = self.cfg["silent"]
        for attr, value in attrs.items():
            obj = doc
            parts = attr.split(".")
            skip = False
            for part in parts[:-1]:
                if hasattr(obj, part):
                    obj = getattr(obj, part)
                else:
                    skip = True
                    if not silent:
                        warnings.warn(Warnings.W116.format(attr=attr))
            if not skip:
                if hasattr(obj, parts[-1]):
                    setattr(obj, parts[-1], value)
                else:
                    if not silent:
                        warnings.warn(Warnings.W116.format(attr=attr))
        return doc

    def to_bytes(self, **kwargs):
        serializers = {
            "cfg": lambda: srsly.json_dumps(self.cfg),
        }
        return util.to_bytes(serializers, [])

    def from_bytes(self, data, **kwargs):
        deserializers = {
            "cfg": lambda b: self.cfg.update(srsly.json_loads(b)),
        }
        util.from_bytes(data, deserializers, [])
        return self

    def to_disk(self, path, **kwargs):
        path = util.ensure_path(path)
        serializers = {
            "cfg": lambda p: srsly.write_json(p, self.cfg),
        }
        return util.to_disk(path, serializers, [])

    def from_disk(self, path, **kwargs):
        path = util.ensure_path(path)
        serializers = {
            "cfg": lambda p: self.cfg.update(srsly.read_json(p)),
        }
        util.from_disk(path, serializers, [])


# Setup backwards compatibility hook for factories
def __getattr__(name):
    if name == "make_doc_cleaner":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_doc_cleaner
    elif name == "make_token_splitter":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_token_splitter
    raise AttributeError(f"module {__name__} has no attribute {name}")



STOP_WORDS = set(
    """
a ak an ankò ant apre ap atò avan avanlè
byen bò byenke

chak

de depi deja deja

e en epi èske

fò fòk

gen genyen

ki kisa kilès kote koukou konsa konbyen konn konnen kounye kouman

la l laa le lè li lye lò

m m' mwen

nan nap nou n'

ou oumenm

pa paske pami pandan pito pou pral preske pwiske

se selman si sou sòt

ta tap tankou te toujou tou tan tout toutotan twòp tèl

w w' wi wè

y y' yo yon yonn

non o oh eh

sa san si swa si

men mèsi oswa osinon

""".split()
)

# Add common contractions, with and without apostrophe variants
contractions = ["m'", "n'", "w'", "y'", "l'", "t'", "k'"]
for apostrophe in ["'", "’", "‘"]:
    for word in contractions:
        STOP_WORDS.add(word.replace("'", apostrophe))



import importlib
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from thinc.api import Config, Model, Optimizer, set_dropout_rate
from thinc.types import Floats2d

from ..errors import Errors
from ..language import Language
from ..scorer import Scorer
from ..tokens import Doc, Span
from ..training import Example
from .spancat import DEFAULT_SPANS_KEY
from .trainable_pipe import TrainablePipe

span_finder_default_config = """
[model]
@architectures = "spacy.SpanFinder.v1"

[model.scorer]
@layers = "spacy.LinearLogistic.v1"
nO = 2

[model.tok2vec]
@architectures = "spacy.Tok2Vec.v2"

[model.tok2vec.embed]
@architectures = "spacy.MultiHashEmbed.v2"
width = 96
rows = [5000, 1000, 2500, 1000]
attrs = ["NORM", "PREFIX", "SUFFIX", "SHAPE"]
include_static_vectors = false

[model.tok2vec.encode]
@architectures = "spacy.MaxoutWindowEncoder.v2"
width = ${model.tok2vec.embed.width}
window_size = 1
maxout_pieces = 3
depth = 4
"""

DEFAULT_SPAN_FINDER_MODEL = Config().from_str(span_finder_default_config)["model"]


def make_span_finder_scorer():
    return span_finder_score


def span_finder_score(examples: Iterable[Example], **kwargs) -> Dict[str, Any]:
    kwargs = dict(kwargs)
    attr_prefix = "spans_"
    key = kwargs["spans_key"]
    kwargs.setdefault("attr", f"{attr_prefix}{key}")
    kwargs.setdefault(
        "getter", lambda doc, key: doc.spans.get(key[len(attr_prefix) :], [])
    )
    kwargs.setdefault("has_annotation", lambda doc: key in doc.spans)
    kwargs.setdefault("allow_overlap", True)
    kwargs.setdefault("labeled", False)
    scores = Scorer.score_spans(examples, **kwargs)
    scores.pop(f"{kwargs['attr']}_per_type", None)
    return scores


def _char_indices(span: Span) -> Tuple[int, int]:
    start = span[0].idx
    end = span[-1].idx + len(span[-1])
    return start, end


class SpanFinder(TrainablePipe):
    """Pipeline that learns span boundaries.

    DOCS: https://spacy.io/api/spanfinder
    """

    def __init__(
        self,
        nlp: Language,
        model: Model[Iterable[Doc], Floats2d],
        name: str = "span_finder",
        *,
        spans_key: str = DEFAULT_SPANS_KEY,
        threshold: float = 0.5,
        max_length: Optional[int] = None,
        min_length: Optional[int] = None,
        scorer: Optional[Callable] = span_finder_score,
    ) -> None:
        """Initialize the span finder.
        model (thinc.api.Model): The Thinc Model powering the pipeline
            component.
        name (str): The component instance name, used to add entries to the
            losses during training.
        threshold (float): Minimum probability to consider a prediction
            positive.
        scorer (Optional[Callable]): The scoring method.
        spans_key (str): Key of the doc.spans dict to save the spans under.
            During initialization and training, the component will look for
            spans on the reference document under the same key.
        max_length (Optional[int]): Maximum length of the produced spans,
            defaults to None meaning unlimited length.
        min_length (Optional[int]): Minimum length of the produced spans,
            defaults to None meaning shortest span length is 1.

        DOCS: https://spacy.io/api/spanfinder#init
        """
        self.vocab = nlp.vocab
        if (max_length is not None and max_length < 1) or (
            min_length is not None and min_length < 1
        ):
            raise ValueError(
                Errors.E1053.format(min_length=min_length, max_length=max_length)
            )
        self.model = model
        self.name = name
        self.scorer = scorer
        self.cfg: Dict[str, Any] = {
            "min_length": min_length,
            "max_length": max_length,
            "threshold": threshold,
            "spans_key": spans_key,
        }

    def predict(self, docs: Iterable[Doc]):
        """Apply the pipeline's model to a batch of docs, without modifying
        them.

        docs (Iterable[Doc]): The documents to predict.
        RETURNS: The models prediction for each document.

        DOCS: https://spacy.io/api/spanfinder#predict
        """
        scores = self.model.predict(docs)
        return scores

    def set_annotations(self, docs: Iterable[Doc], scores: Floats2d) -> None:
        """Modify a batch of Doc objects, using pre-computed scores.
        docs (Iterable[Doc]): The documents to modify.
        scores: The scores to set, produced by SpanFinder predict method.

        DOCS: https://spacy.io/api/spanfinder#set_annotations
        """
        offset = 0
        for i, doc in enumerate(docs):
            doc.spans[self.cfg["spans_key"]] = []
            starts = []
            ends = []
            doc_scores = scores[offset : offset + len(doc)]

            for token, token_score in zip(doc, doc_scores):
                if token_score[0] >= self.cfg["threshold"]:
                    starts.append(token.i)
                if token_score[1] >= self.cfg["threshold"]:
                    ends.append(token.i)

            for start in starts:
                for end in ends:
                    span_length = end + 1 - start
                    if span_length < 1:
                        continue
                    if (
                        self.cfg["min_length"] is None
                        or self.cfg["min_length"] <= span_length
                    ) and (
                        self.cfg["max_length"] is None
                        or span_length <= self.cfg["max_length"]
                    ):
                        doc.spans[self.cfg["spans_key"]].append(doc[start : end + 1])
            offset += len(doc)

    def update(
        self,
        examples: Iterable[Example],
        *,
        drop: float = 0.0,
        sgd: Optional[Optimizer] = None,
        losses: Optional[Dict[str, float]] = None,
    ) -> Dict[str, float]:
        """Learn from a batch of documents and gold-standard information,
        updating the pipe's model. Delegates to predict and get_loss.
        examples (Iterable[Example]): A batch of Example objects.
        drop (float): The dropout rate.
        sgd (Optional[thinc.api.Optimizer]): The optimizer.
        losses (Optional[Dict[str, float]]): Optional record of the loss during
            training. Updated using the component name as the key.
        RETURNS (Dict[str, float]): The updated losses dictionary.

        DOCS: https://spacy.io/api/spanfinder#update
        """
        if losses is None:
            losses = {}
        losses.setdefault(self.name, 0.0)
        predicted = [eg.predicted for eg in examples]
        set_dropout_rate(self.model, drop)
        scores, backprop_scores = self.model.begin_update(predicted)
        loss, d_scores = self.get_loss(examples, scores)
        backprop_scores(d_scores)
        if sgd is not None:
            self.finish_update(sgd)
        losses[self.name] += loss
        return losses

    def get_loss(self, examples, scores) -> Tuple[float, Floats2d]:
        """Find the loss and gradient of loss for the batch of documents and
        their predicted scores.
        examples (Iterable[Examples]): The batch of examples.
        scores: Scores representing the model's predictions.
        RETURNS (Tuple[float, Floats2d]): The loss and the gradient.

        DOCS: https://spacy.io/api/spanfinder#get_loss
        """
        truths, masks = self._get_aligned_truth_scores(examples, self.model.ops)
        d_scores = scores - self.model.ops.asarray2f(truths)
        d_scores *= masks
        loss = float((d_scores**2).sum())
        return loss, d_scores

    def _get_aligned_truth_scores(self, examples, ops) -> Tuple[Floats2d, Floats2d]:
        """Align scores of the predictions to the references for calculating
        the loss.
        """
        truths = []
        masks = []
        for eg in examples:
            if eg.x.text != eg.y.text:
                raise ValueError(Errors.E1054.format(component="span_finder"))
            n_tokens = len(eg.predicted)
            truth = ops.xp.zeros((n_tokens, 2), dtype="float32")
            mask = ops.xp.ones((n_tokens, 2), dtype="float32")
            if self.cfg["spans_key"] in eg.reference.spans:
                for span in eg.reference.spans[self.cfg["spans_key"]]:
                    ref_start_char, ref_end_char = _char_indices(span)
                    pred_span = eg.predicted.char_span(
                        ref_start_char, ref_end_char, alignment_mode="expand"
                    )
                    pred_start_char, pred_end_char = _char_indices(pred_span)
                    start_match = pred_start_char == ref_start_char
                    end_match = pred_end_char == ref_end_char
                    if start_match:
                        truth[pred_span[0].i, 0] = 1
                    else:
                        mask[pred_span[0].i, 0] = 0
                    if end_match:
                        truth[pred_span[-1].i, 1] = 1
                    else:
                        mask[pred_span[-1].i, 1] = 0
            truths.append(truth)
            masks.append(mask)
        truths = ops.xp.concatenate(truths, axis=0)
        masks = ops.xp.concatenate(masks, axis=0)
        return truths, masks

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Optional[Language] = None,
    ) -> None:
        """Initialize the pipe for training, using a representative set
        of data examples.
        get_examples (Callable[[], Iterable[Example]]): Function that
            returns a representative sample of gold-standard Example objects.
        nlp (Optional[Language]): The current nlp object the component is part
            of.

        DOCS: https://spacy.io/api/spanfinder#initialize
        """
        subbatch: List[Example] = []

        for eg in get_examples():
            if len(subbatch) < 10:
                subbatch.append(eg)

        if subbatch:
            docs = [eg.reference for eg in subbatch]
            Y, _ = self._get_aligned_truth_scores(subbatch, self.model.ops)
            self.model.initialize(X=docs, Y=Y)
        else:
            self.model.initialize()


# Setup backwards compatibility hook for factories
def __getattr__(name):
    if name == "make_span_finder":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_span_finder
    raise AttributeError(f"module {__name__} has no attribute {name}")



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
)

from thinc.api import Model, Optimizer

from .compat import Protocol, runtime_checkable

if TYPE_CHECKING:
    from .language import Language
    from .training import Example


@runtime_checkable
class TrainableComponent(Protocol):
    model: Any
    is_trainable: bool

    def update(
        self,
        examples: Iterable["Example"],
        *,
        drop: float = 0.0,
        sgd: Optional[Optimizer] = None,
        losses: Optional[Dict[str, float]] = None,
    ) -> Dict[str, float]: ...

    def finish_update(self, sgd: Optimizer) -> None: ...


@runtime_checkable
class InitializableComponent(Protocol):
    def initialize(
        self,
        get_examples: Callable[[], Iterable["Example"]],
        nlp: "Language",
        **kwargs: Any,
    ): ...


@runtime_checkable
class ListenedToComponent(Protocol):
    model: Any
    listeners: Sequence[Model]
    listener_map: Dict[str, Sequence[Model]]
    listening_components: List[str]

    def add_listener(self, listener: Model, component_name: str) -> None: ...

    def remove_listener(self, listener: Model, component_name: str) -> bool: ...

    def find_listeners(self, component) -> None: ...



import importlib.util
import shutil
import sys
from typing import Optional, Sequence
from urllib.parse import urljoin

import requests
import typer
from wasabi import msg

from .. import about
from ..errors import OLD_MODEL_SHORTCUTS
from ..util import (
    get_minor_version,
    is_in_interactive,
    is_in_jupyter,
    is_package,
    is_prerelease_version,
    run_command,
)
from ._util import SDIST_SUFFIX, WHEEL_SUFFIX, Arg, Opt, app


@app.command(
    "download",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def download_cli(
    # fmt: off
    ctx: typer.Context,
    model: str = Arg(..., help="Name of pipeline package to download"),
    direct: bool = Opt(
        False, "--direct", "-d", "-D", help="Force direct download of name + version"
    ),
    sdist: bool = Opt(
        False,
        "--sdist",
        "-S",
        help="Download sdist (.tar.gz) archive instead of pre-built binary wheel",
    ),
    url: str = Opt(None, "--url", "-U", help="Download from given url"),
    # fmt: on
):
    """
    Download compatible trained pipeline from the default download path using
    pip. If --direct flag is set, the command expects the full package name with
    version. For direct downloads, the compatibility check will be skipped. All
    additional arguments provided to this command will be passed to `pip install`
    on package installation.

    DOCS: https://spacy.io/api/cli#download
    AVAILABLE PACKAGES: https://spacy.io/models
    """
    download(model, direct, sdist, url, *ctx.args)


def download(
    model: str,
    direct: bool = False,
    sdist: bool = False,
    custom_url: Optional[str] = None,
    *pip_args,
) -> None:
    if (
        not (is_package("spacy") or is_package("spacy-nightly"))
        and "--no-deps" not in pip_args
    ):
        msg.warn(
            "Skipping pipeline package dependencies and setting `--no-deps`. "
            "You don't seem to have the spaCy package itself installed "
            "(maybe because you've built from source?), so installing the "
            "package dependencies would cause spaCy to be downloaded, which "
            "probably isn't what you want. If the pipeline package has other "
            "dependencies, you'll have to install them manually."
        )
        pip_args = pip_args + ("--no-deps",)
    if direct:
        # Reject model names with '/', in order to prevent shenanigans.
        if "/" in model:
            msg.fail(
                title="Model download rejected",
                text=f"Cannot download model '{model}'. Models are expected to be file names, not URLs or fragments",
                exits=True,
            )
        components = model.split("-")
        model_name = "".join(components[:-1])
        version = components[-1]
    else:
        model_name = model
        if model in OLD_MODEL_SHORTCUTS:
            msg.warn(
                f"As of spaCy v3.0, shortcuts like '{model}' are deprecated. Please "
                f"use the full pipeline package name '{OLD_MODEL_SHORTCUTS[model]}' instead."
            )
            model_name = OLD_MODEL_SHORTCUTS[model]
        compatibility = get_compatibility()
        version = get_version(model_name, compatibility)

    filename = get_model_filename(model_name, version, sdist)

    download_model(filename, pip_args, custom_url)
    msg.good(
        "Download and installation successful",
        f"You can now load the package via spacy.load('{model_name}')",
    )
    if is_in_jupyter():
        reload_deps_msg = (
            "If you are in a Jupyter or Colab notebook, you may need to "
            "restart Python in order to load all the package's dependencies. "
            "You can do this by selecting the 'Restart kernel' or 'Restart "
            "runtime' option."
        )
        msg.warn(
            "Restart to reload dependencies",
            reload_deps_msg,
        )
    elif is_in_interactive():
        reload_deps_msg = (
            "If you are in an interactive Python session, you may need to "
            "exit and restart Python to load all the package's dependencies. "
            "You can exit with Ctrl-D (or Ctrl-Z and Enter on Windows)."
        )
        msg.warn(
            "Restart to reload dependencies",
            reload_deps_msg,
        )


def get_model_filename(model_name: str, version: str, sdist: bool = False) -> str:
    dl_tpl = "{m}-{v}/{m}-{v}{s}"
    suffix = SDIST_SUFFIX if sdist else WHEEL_SUFFIX
    filename = dl_tpl.format(m=model_name, v=version, s=suffix)
    return filename


def get_compatibility() -> dict:
    if is_prerelease_version(about.__version__):
        version: Optional[str] = about.__version__
    else:
        version = get_minor_version(about.__version__)
    r = requests.get(about.__compatibility__)
    if r.status_code != 200:
        msg.fail(
            f"Server error ({r.status_code})",
            f"Couldn't fetch compatibility table. Please find a package for your spaCy "
            f"installation (v{about.__version__}), and download it manually. "
            f"For more details, see the documentation: "
            f"https://spacy.io/usage/models",
            exits=1,
        )
    comp_table = r.json()
    comp = comp_table["spacy"]
    if version not in comp:
        msg.fail(f"No compatible packages found for v{version} of spaCy", exits=1)
    return comp[version]


def get_version(model: str, comp: dict) -> str:
    if model not in comp:
        msg.fail(
            f"No compatible package found for '{model}' (spaCy v{about.__version__})",
            exits=1,
        )
    return comp[model][0]


def get_latest_version(model: str) -> str:
    comp = get_compatibility()
    return get_version(model, comp)


def download_model(
    filename: str,
    user_pip_args: Optional[Sequence[str]] = None,
    custom_url: Optional[str] = None,
) -> None:
    # Construct the download URL carefully. We need to make sure we don't
    # allow relative paths or other shenanigans to trick us into download
    # from outside our own repo.
    base_url = custom_url if custom_url else about.__download_url__
    # urljoin requires that the path ends with /, or the last path part will be dropped
    if not base_url.endswith("/"):
        base_url = about.__download_url__ + "/"
    download_url = urljoin(base_url, filename)
    if not download_url.startswith(about.__download_url__):
        raise ValueError(f"Download from {filename} rejected. Was it a relative path?")
    pip_args = list(user_pip_args) if user_pip_args is not None else []
    cmd = _get_pip_install_cmd() + pip_args + [download_url]
    run_command(cmd)


def _get_pip_install_cmd() -> list:
    if importlib.util.find_spec("pip") is not None:
        return [sys.executable, "-m", "pip", "install"]
    elif shutil.which("uv"):
        return ["uv", "pip", "install"]
    else:
        msg.fail(
            "No package installer found",
            "spaCy requires either pip or uv to download models. "
            "Please install one of them and try again.",
            exits=1,
        )



from typing import Iterator, Tuple, Union

from ...errors import Errors
from ...symbols import NOUN, PRON, PROPN
from ...tokens import Doc, Span


def noun_chunks(doclike: Union[Doc, Span]) -> Iterator[Tuple[int, int, int]]:
    """
    Detect base noun phrases from a dependency parse for Haitian Creole.
    Works on both Doc and Span objects.
    """

    # Core nominal dependencies common in Haitian Creole
    labels = [
        "nsubj",
        "obj",
        "obl",
        "nmod",
        "appos",
        "ROOT",
    ]

    # Modifiers to optionally include in chunk (to the right)
    post_modifiers = ["compound", "flat", "flat:name", "fixed"]

    doc = doclike.doc
    if not doc.has_annotation("DEP"):
        raise ValueError(Errors.E029)

    np_deps = {doc.vocab.strings.add(label) for label in labels}
    np_mods = {doc.vocab.strings.add(mod) for mod in post_modifiers}
    conj_label = doc.vocab.strings.add("conj")
    np_label = doc.vocab.strings.add("NP")
    adp_pos = doc.vocab.strings.add("ADP")
    cc_pos = doc.vocab.strings.add("CCONJ")

    prev_end = -1
    for i, word in enumerate(doclike):
        if word.pos not in (NOUN, PROPN, PRON):
            continue
        if word.left_edge.i <= prev_end:
            continue

        if word.dep in np_deps:
            right_end = word
            # expand to include known modifiers to the right
            for child in word.rights:
                if child.dep in np_mods:
                    right_end = child.right_edge
                elif child.pos == NOUN:
                    right_end = child.right_edge

            left_index = word.left_edge.i
            # Skip prepositions at the start
            if word.left_edge.pos == adp_pos:
                left_index += 1

            prev_end = right_end.i
            yield left_index, right_end.i + 1, np_label

        elif word.dep == conj_label:
            head = word.head
            while head.dep == conj_label and head.head.i < head.i:
                head = head.head
            if head.dep in np_deps:
                left_index = word.left_edge.i
                if word.left_edge.pos == cc_pos:
                    left_index += 1
                prev_end = word.i
                yield left_index, word.i + 1, np_label


SYNTAX_ITERATORS = {"noun_chunks": noun_chunks}



from typing import List, Tuple

from ...pipeline import Lemmatizer
from ...tokens import Token


class HaitianCreoleLemmatizer(Lemmatizer):
    """
    Minimal Haitian Creole lemmatizer.
    Returns a word's base form based on rules and lookup,
    or defaults to the original form.
    """

    def is_base_form(self, token: Token) -> bool:
        morph = token.morph.to_dict()
        upos = token.pos_.lower()

        # Consider unmarked forms to be base
        if upos in {"noun", "verb", "adj", "adv"}:
            if not morph:
                return True
            if upos == "noun" and morph.get("Number") == "Sing":
                return True
            if upos == "verb" and morph.get("VerbForm") == "Inf":
                return True
            if upos == "adj" and morph.get("Degree") == "Pos":
                return True
        return False

    def rule_lemmatize(self, token: Token) -> List[str]:
        string = token.text.lower()
        pos = token.pos_.lower()
        cache_key = (token.orth, token.pos)
        if cache_key in self.cache:
            return self.cache[cache_key]

        forms = []

        # fallback rule: just return lowercased form
        forms.append(string)

        self.cache[cache_key] = forms
        return forms

    @classmethod
    def get_lookups_config(cls, mode: str) -> Tuple[List[str], List[str]]:
        if mode == "rule":
            required = ["lemma_lookup", "lemma_rules", "lemma_exc", "lemma_index"]
            return (required, [])
        return super().get_lookups_config(mode)



from ...symbols import NORM, ORTH
from ...util import update_exc
from ..tokenizer_exceptions import BASE_EXCEPTIONS

"""
 All rules and exceptions were taken from the "Gaelic Orthographic Conventions
of 2009" (GOC) and from the "Annotated Reference Corpus of Scottish Gaelic" (ARCOSG). I did
my best to ensure this tokenizer would lead to text as close as possible to the
tokenization of the ARCOSG and the conventions in the GOC.


ARCOSG: https://github.com/Gaelic-Algorithmic-Research-Group/ARCOSG
GOC: https://www.gaidhlig.scot/wp-content/uploads/2021/03/GOC-2009-English.pdf
"""

# Compound words
_exc = {
    "càil": [{ORTH: "cà", NORM: "càite"}, {ORTH: "il", NORM: "bheil"}],
    "sna": [{ORTH: "s", NORM: "anns"}, {ORTH: "na", NORM: "na"}],
    "orra": [{ORTH: "orr", NORM: "air"}, {ORTH: "a", NORM: "do"}],
    "fiùs": [{ORTH: "fiù", NORM: "fiù"}, {ORTH: "s", NORM: "'s"}],
}


# Hyphenations that are alternative forms of words
for exc_data in [
    {ORTH: "fa-near", NORM: "fainear"},
    {ORTH: "Fa-near", NORM: "Fainear"},
]:
    _exc[exc_data[ORTH]] = [exc_data]


# Abreviations and shortened words
for exc_data in [
    {ORTH: "'", NORM: "a"},
    {ORTH: "'S", NORM: "Agus"},
    {ORTH: "'s", NORM: "agus"},
    {ORTH: "B'", NORM: "Bu"},
    {ORTH: "b'", NORM: "bu"},
    {ORTH: "D'", NORM: "Do"},
    {ORTH: "d'", NORM: "do"},
    {ORTH: "'M", NORM: "Am"},
    {ORTH: "'m", NORM: "am"},
    {ORTH: "M'", NORM: "Mo"},
    {ORTH: "m'", NORM: "mo"},
    {ORTH: "'n", NORM: "an"},
    {ORTH: "'N", NORM: "An"},
    {ORTH: "Th'", NORM: "Tha"},
    {ORTH: "th'", NORM: "tha"},
]:
    _exc[exc_data[ORTH]] = [exc_data]


# Words with a leading apostrophe
for orth in """
  'ac
  'Ac
  'ad
  'Ad
  'ar
  'Ar
  'bhuannachd
  'Bhuannachd
  'd
  'D
  'eil
  'Eil
  'eug
  'Eug
  'g
  'G
  'ga
  'Ga
  'gad
  'Gad
  'gam
  'Gam
  'gan
  'Gan
  'gar
  'Gar
  'gur
  'Gur
  'ic
  'Ic
  'il
  'Il
  'ill'
  'Ill'
  'ille
  'Ille
  'illean
  'Illean
  'iodh
  'Iodh
  'l
  'L
  'm
  'M
  'n
  'N
  'na
  'Na
  'nad
  'Nad
  'nam
  'Nam
  'nan
  'Nan
  'nar
  'Nar
  'neil
  'Neil
  'nise
  'Nise
  'nuair
  'Nuair
  'nur
  'Nur
  's
  'S
  'sa
  'Sa
  'sa'
  'Sa'
  'san
  'San
  'sann
  'Sann
  'se
  'Se
  'sna
  'Sna
  'son
  'Son
  'urchaidh
  'Urchaidh
  """.split():
    _exc[orth] = [{ORTH: orth}]

# Words with a trailing or middling apostrophe
for orth in """
  a'
  A'
  a'd
  A'd
  a'm
  A'm
  a's
  A's
  ac'
  Ac'
  agads'
  Agads'
  agams'
  Agams'
  aig'
  Aig'
  annams'
  Annams'
  ars'
  Ars'
  b'
  B'
  ball'
  Ball'
  bioraicht'
  Bioraicht'
  bh'
  Bh'
  bhail'
  Bhail'
  bhall'
  Bhall'
  bheath'
  Bheath'
  bhliadhn'
  Bhliadhn'
  bliadhn'
  Bliadhn'
  bonnant'
  Bonnant'
  brist'
  Brist'
  bàt'
  Bàt'
  c'à
  C'à
  camp'
  Camp'
  chalp'
  Chalp'
  champ'
  Champ'
  chomhairl'
  Chomhairl'
  chual'
  Chual'
  chuimhn'
  Chuimhn'
  colaisd'
  Colaisd'
  comhl'
  Comhl'
  comhairl'
  Comhairl'
  creids'
  Creids'
  cual'
  Cual'
  cuimhn'
  Cuimhn'
  cuin'
  Cuin'
  cà'
  Cà'
  càit'
  Càit'
  d'
  D'
  d'readh
  D'readh
  d'reaghadh
  D'reaghadh
  daoin'
  Daoin'
  deimhinn'
  Deimhinn'
  de'n
  De'n
  dh'
  Dh'
  dhaib'
  Dhaib'
  dhaoin'
  Dhaoin'
  dhòmhs'
  Dhòmhs'
  dhu'sa
  Dhu'sa
  dhuin'
  Dhuin'
  do'n
  Do'n
  duin'
  Duin'
  dòch'
  Dòch'
  dùint'
  Dùint'
  eil'
  Eil'
  f'a
  F'a
  fac'
  Fac'
  fad'
  Fad'
  fhac'
  Fhac'
  fhad'
  Fhad'
  fhaid'
  Fhaid'
  fhaisg'
  Fhaisg'
  fhiosd'
  Fhiosd'
  fàilt'
  Fàilt'
  g'
  G'
  gàir'
  Gàir'
  ghill'
  Ghill'
  gill'
  Gill'
  inns'
  Inns'
  innt'
  Innt'
  ionnsaicht'
  Ionnsaicht'
  leams'
  Leams'
  leoth'
  Leoth'
  lobht'
  Lobht'
  m'
  M'
  m'a
  M'a
  m's
  M's
  mhuth'
  Mhuth'
  mhòr'
  Mhòr'
  mis'
  Mis'
  mu'n
  Mu'n
  mòr'
  Mòr'
  oirr'
  Oirr'
  o'n
  O'n
  phàp'
  Phàp'
  pàp'
  Pàp'
  pòs'
  Pòs'
  prionns'
  Prionns'
  r'
  R'
  riums'
  Riums'
  riuth'
  Riuth'
  ro'n
  Ro'n
  sa'
  Sa'
  sgoil'
  Sgoil'
  sgìr'
  Sgìr'
  sheòrs'
  Sheòrs'
  sin'
  Sin'
  stall'
  Stall'
  sìod'
  Sìod'
  sònraicht'
  Sònraicht'
  t'
  T'
  taigh'
  Taigh'
  tein'
  Tein'
  teoth'
  Teoth'
  th'
  Th'
  thoilicht'
  Thoilicht'
  thuc'
  Thuc'
  thuigs'
  Thuigs'
  thus'
  Thus'
  thàna'
  Thàna'
  toilicht'
  Toilicht'
  tro'
  Tro'
  uisg'
  Uisg'
  àit'
  Àit'
  òg'
  Òg'
  """.split():
    _exc[orth] = [{ORTH: orth}]


# Hyphenations that should remain as single tokens
for orth in """
'n-dràsda
'N-dràsda
-bhliadhn'
-bhliadhn'
a-bhos
A-bhos
a-bhòn-dè
A-bhòn-dè
a-cheart
A-cheart
a-chèile
A-chèile
a-deas
A-deas
a-mach
A-mach
a-mhàin
A-mhàin
a-muigh
A-muigh
a-màireach
A-màireach
a-nall
A-nall
a-neist
A-neist
a-ni
A-ni
a-nis
A-nis
a-nisd
A-nisd
a-nise
A-nise
a-nist
A-nist
a-niste
A-niste
a-nochd
A-nochd
a-nuas
A-nuas
a-null
A-null
a-raoir
A-raoir
a-riamh
A-riamh
a-rithist
A-rithist
a-rèiste
A-rèiste
a-rìs
A-rìs
a-seo
A-seo
a-sin
A-sin
a-sineach
A-sineach
a-siud
A-siud
a-staigh
A-staigh
a-steach
A-steach
a-tuath
A-tuath
aca-san
Aca-san
agad-sa
Agad-sa
agam-sa
Agam-sa
aghaidh-bhualaich
Aghaidh-bhualaich
aice-se
Aice-se
aige-san
Aige-san
ainmeannan-àite
Ainmeannan-àite
air-san
Air-san
am-bliadhna
Am-bliadhna
am-màireach
Am-màireach
amp-head
Amp-head
an-diugh
An-diugh
an-dràsd
An-dràsd
an-dràsda
An-dràsda
an-dràst
An-dràst
an-dràsta
An-dràsta
an-dè
An-dè
an-dé
An-dé
an-nise
An-nise
an-nochd
An-nochd
an-raoir
An-raoir
an-uiridh
An-uiridh
an-àbhaisteach
An-àbhaisteach
an-àird
An-àird
an-àirde
An-àirde
an-àrda
An-àrda
ana-ceartas
Ana-ceartas
ana-seo
Ana-seo
ana-sin
Ana-sin
ana-siud
Ana-siud
annam-s'
Annam-s'
ao-coltach
Ao-coltach
aobhar-sa
Aobhar-sa
aois-léinn
Aois-léinn
aona-ghnothaich
Aona-ghnothaich
ar-a-mach
Ar-a-mach
ard-easbaig
Ard-easbaig
ard-luchd-poilitics
Ard-luchd-poilitics
ath-bhaile
Ath-bhaile
ath-bheòthachadh
Ath-bheòthachadh
ath-bhliadhna
Ath-bhliadhna
ath-ghiollachd
Ath-ghiollachd
ath-nuadhais
Ath-nuadhais
ath-sgrùdadh
Ath-sgrùdadh
ath-thriop
Ath-thriop
athair-san
Athair-san
baile-ciùird
Baile-ciùird
ball-coise
Ball-coise
ball-pàrlamaid
Ball-pàrlamaid
ball-sampaill
Ball-sampaill
balla-mara
Balla-mara
ban-chompanach
Ban-chompanach
ban-fhuamhaire
Ban-fhuamhaire
ban-ghillìosach
Ban-ghillìosach
ban-righ'nn
Ban-righ'nn
ban-rìgh
Ban-rìgh
bana-bhàird
Bana-bhàird
bana-chompanaich
Bana-chompanaich
bana-phòsda
Bana-phòsda
banas-taighe
Banas-taighe
beairt-fhuaigheil
Beairt-fhuaigheil
beairt-fuaigheil
Beairt-fuaigheil
bean-gairm
Bean-gairm
bean-phòsta
Bean-phòsta
bean-taighe
Bean-taighe
beul-aithris
Beul-aithris
beò-shlàint
Beò-shlàint
beò-shlàint'
Beò-shlàint'
beò-shlàinte
Beò-shlàinte
bhaga-sgoil
Bhaga-sgoil
bhall-pàrlamaid
Bhall-pàrlamaid
bhan-chompanach
Bhan-chompanach
bhan-dòmhnallach
Bhan-dòmhnallach
bhan-phrionnsa
Bhan-phrionnsa
bhan-righinn
Bhan-righinn
bhan-sheinneadair
Bhan-sheinneadair
bharr-iall
Bharr-iall
bhata-làidir
Bhata-làidir
bhath-room
Bhath-room
bheachd-sa
Bheachd-sa
bheachd-san
Bheachd-san
bheairt-fhighe
Bheairt-fhighe
bheairtean-fuaigheil
Bheairtean-fuaigheil
bheinn-sheilg
Bheinn-sheilg
bheul-aithris
Bheul-aithris
bheò-ghlacadh
Bheò-ghlacadh
bhith-beò
Bhith-beò
bhithinn-sa
Bhithinn-sa
bhogsa-chiùil
Bhogsa-chiùil
bhonn-stéidh
Bhonn-stéidh
bhràithrean-sa
Bhràithrean-sa
bhuain-mhòine
Bhuain-mhòine
bhun-sheòrsa
Bhun-sheòrsa
bhàn-righinn
Bhàn-righinn
bhàn-rinn
Bhàn-rinn
bhàn-rìgh
Bhàn-rìgh
bhàta-aiseig
Bhàta-aiseig
bhàta-sa
Bhàta-sa
bird-watcher
Bird-watcher
bith-beò
Bith-beò
bithinn-sa
Bithinn-sa
bliadhna-sa
Bliadhna-sa
bogha-saighead
Bogha-saighead
boma-peatroil
Boma-peatroil
bristeadh-a-mach
Bristeadh-a-mach
buidhean-cathrannais
Buidhean-cathrannais
buille-a-mach
Buille-a-mach
buille-shaor
Buille-shaor
bun-adhbharan
Bun-adhbharan
bun-chomharraidhean
Bun-chomharraidhean
bun-fhiosrachadh
Bun-fhiosrachadh
bun-sgoil
Bun-sgoil
bun-stèidh
Bun-stèidh
bàt-aiseig
Bàt-aiseig
bàta-aiseig
Bàta-aiseig
bàta-bathair
Bàta-bathair
cainnt-san
Cainnt-san
cal-mac
Cal-mac
carraighean-cuimhne
Carraighean-cuimhne
cead-telebhisean
Cead-telebhisean
ceann-cinnidh
Ceann-cinnidh
ceann-suidhe
Ceann-suidhe
chanain-sa
Chanain-sa
chaolas-arcach
Chaolas-arcach
charge-adh
Charge-adh
cheala-deug
Cheala-deug
chealla-deug
Chealla-deug
cheann-cinnidh
Cheann-cinnidh
cheann-feadhna
Cheann-feadhna
cheann-suidhe
Cheann-suidhe
chearc-fhraoich
Chearc-fhraoich
chearcall-meadhain
Chearcall-meadhain
chearcall-mheadhain
Chearcall-mheadhain
chlann-nighean
Chlann-nighean
chlàr-ama
Chlàr-ama
chlò-bhuaileadh
Chlò-bhuaileadh
chlò-bhualadh
Chlò-bhualadh
cho-chreutairean
Cho-chreutairean
cho-dhùin
Cho-dhùin
cho-dhùnadh
Cho-dhùnadh
cho-dhùnaidhean
Cho-dhùnaidhean
cho-fhaireachdainn
Cho-fhaireachdainn
cho-labhairt
Cho-labhairt
cho-obraiche
Cho-obraiche
cho-roinn
Cho-roinn
chom-pàirt
Chom-pàirt
chorra-ghritheach
Chorra-ghritheach
chrann-snàth
Chrann-snàth
chreach-s'
Chreach-s'
chrith-thalmhainn
Chrith-thalmhainn
chàch-a-chéile
Chàch-a-chéile
cinn-chuspair
Cinn-chuspair
cinn-iùil
Cinn-iùil
cion-doighe
Cion-doighe
clachan-meallain
Clachan-meallain
clann-sgoile
Clann-sgoile
claon-fhaireachdainn
Claon-fhaireachdainn
claon-shamhail
Claon-shamhail
cluicheadairean-meadhain
Cluicheadairean-meadhain
clàran-ama
Clàran-ama
cléir-seanchain
Cléir-seanchain
clò-bhualadair
Clò-bhualadair
clò-bhualadh
Clò-bhualadh
co-aimsireach
Co-aimsireach
co-bhanntachd
Co-bhanntachd
co-bhuannachd
Co-bhuannachd
co-buannachd
Co-buannachd
co-cheangail
Co-cheangail
co-cheangailte
Co-cheangailte
co-cheangal
Co-cheangal
co-chreutairean
Co-chreutairean
co-chruinneachadh
Co-chruinneachadh
co-dhiu
Co-dhiu
co-dhiubh
Co-dhiubh
co-dhiù
Co-dhiù
co-dhiùbh
Co-dhiùbh
co-dhùnadh
Co-dhùnadh
co-dhùnaidhean
Co-dhùnaidhean
co-fhaireachadh
Co-fhaireachadh
co-fhaireachdainn
Co-fhaireachdainn
co-impirean
Co-impirean
co-ionad
Co-ionad
co-ionann
Co-ionann
co-labhairt
Co-labhairt
co-labhairtean
Co-labhairtean
co-obrachadh
Co-obrachadh
co-sheirm
Co-sheirm
co-theacs
Co-theacs
coimeas-meudachd
Coimeas-meudachd
cola-deug
Cola-deug
com-pàirt
Com-pàirt
cope-adh
Cope-adh
crann-aodaich
Crann-aodaich
crann-snàth
Crann-snàth
crann-tarsainn
Crann-tarsainn
craobh-sgaoileadh
Craobh-sgaoileadh
crith-thalmhainn
Crith-thalmhainn
cruth-rannsachadh
Cruth-rannsachadh
cuid-eigin
Cuid-eigin
cumail-san
Cumail-san
cur-gu-buil
Cur-gu-buil
cur-seachad
Cur-seachad
cur-seachadan
Cur-seachadan
cìs-comhairle
Cìs-comhairle
cò-dhunadh
Cò-dhunadh
còmhlan-ciùil
Còmhlan-ciùil
cùis-lagh
Cùis-lagh
cùl-chàineadh
Cùl-chàineadh
cùl-shleamhnach
Cùl-shleamhnach
cùl-taic
Cùl-taic
da-rìreabh
Da-rìreabh
da-rìreadh
Da-rìreadh
da-rìribh
Da-rìribh
deagh-ghean
Deagh-ghean
dearg-fhuileach
Dearg-fhuileach
deireadh-sheachdain
Deireadh-sheachdain
deoch-làidir
Deoch-làidir
dha-rìreabh
Dha-rìreabh
dha-rìribh
Dha-rìribh
dhaibh-san
Dhaibh-san
dhe-salin-adh
Dhe-salin-adh
dhe-salt-adh
Dhe-salt-adh
dheidhinn-sa
Dheidhinn-sa
dhol-sìos
Dhol-sìos
dhomh-s'
Dhomh-s'
dhuine-dubh
Dhuine-dubh
dhà-san
Dhà-san
dhòigh-beatha
Dhòigh-beatha
di-sathairne
Di-sathairne
dian-amharc
Dian-amharc
dlùth-cheangal
Dlùth-cheangal
do-chreidsinneach
Do-chreidsinneach
do-labhairt
Do-labhairt
do-sheachant'
Do-sheachant'
dol-a-mach
Dol-a-mach
dol-air-adhart
Dol-air-adhart
dubh-chàineadh
Dubh-chàineadh
dubh-ghorm
Dubh-ghorm
dà-chultarach
Dà-chultarach
dà-reug
Dà-reug
dàn-mòr
Dàn-mòr
dì-moladh
Dì-moladh
dòigh-beatha
Dòigh-beatha
dòighean-beatha
Dòighean-beatha
e-mail
E-mail
eadar-dhealachadh
Eadar-dhealachadh
eadar-dhealachaidhean
Eadar-dhealachaidhean
eadar-dhealaichte
Eadar-dhealaichte
eadar-nàiseanta
Eadar-nàiseanta
earbainn-s
Earbainn-s
eàrr-ràdh
Eàrr-ràdh
eòrp-innseanach
Eòrp-innseanach
fa-leth
Fa-leth
fa-near
Fa-near
fad-as
Fad-as
fad-thréimhseach
Fad-thréimhseach
feadaig-mhonaidh
Feadaig-mhonaidh
fealla-dhà
Fealla-dhà
fear-a-ropa
Fear-a-ropa
fear-ceasnachaidh
Fear-ceasnachaidh
fear-faire
Fear-faire
fear-gairm
Fear-gairm
fear-glèidhidh
Fear-glèidhidh
fear-labhairt
Fear-labhairt
fear-naidheachd
Fear-naidheachd
fear-pòsta
Fear-pòsta
fear-sgrùdaidh
Fear-sgrùdaidh
fear-teagaisg
Fear-teagaisg
fear-trèinidh
Fear-trèinidh
fear-éisteachd
Fear-éisteachd
feed-adh
Feed-adh
fhear-ghlèidhidh
Fhear-ghlèidhidh
fhear-gleidhidh
Fhear-gleidhidh
fhear-glèidhidh
Fhear-glèidhidh
fhear-labhairt
Fhear-labhairt
fhear-leughaidh
Fhear-leughaidh
fhear-sa
Fhear-sa
fhear-sgrùdaidh
Fhear-sgrùdaidh
fhir-cinnidh
Fhir-cinnidh
fhéin-ìomhaigh
Fhéin-ìomhaigh
fhìor-luachmhor
Fhìor-luachmhor
fois-fhòirneirt
Fois-fhòirneirt
fàs-bheairtean
Fàs-bheairtean
féin-mhisneachd
Féin-mhisneachd
féin-mholadh
Féin-mholadh
fìor-thàbhachdach
Fìor-thàbhachdach
ge-ta
Ge-ta
ge-tà
Ge-tà
ged-tà
Ged-tà
geàrr-chunntais
Geàrr-chunntais
geàrr-chunntas
Geàrr-chunntas
geàrr-thréimhseach
Geàrr-thréimhseach
ghuth-thàmh
Ghuth-thàmh
glain'-amhairc
Glain'-amhairc
glas-ghuib
Glas-ghuib
gnàth-bhriathrachas
Gnàth-bhriathrachas
gàrradh-crìche
Gàrradh-crìche
h-
H-
h-ana-miannaibh
H-ana-miannaibh
h-uile
H-uile
hó-ró
Hó-ró
iar-mhinistear
Iar-mhinistear
inneal-spreadhaidh
Inneal-spreadhaidh
ionad-còmhnaidh
Ionad-còmhnaidh
join-adh
Join-adh
latha-an-diugh
Latha-an-diugh
leam-sa
Leam-sa
leas-adh
Leas-adh
lease-adh
Lease-adh
leat-sa
Leat-sa
leotha-san
Leotha-san
leth-char
Leth-char
leth-cheud
Leth-cheud
leth-ghàidhealtachd
Leth-ghàidhealtachd
leth-pocannan
Leth-pocannan
leth-sgeulan
Leth-sgeulan
leth-uair
Leth-uair
leughadh-ne
Leughadh-ne
lighiche-sprèidh
Lighiche-sprèidh
linn-an-òir
Linn-an-òir
litir-aonta
Litir-aonta
loma-làn
Loma-làn
lost-s'
Lost-s'
luchd-altram
Luchd-altram
luchd-altruim
Luchd-altruim
luchd-amhairc
Luchd-amhairc
luchd-ciùil
Luchd-ciùil
luchd-cruinneachaidh
Luchd-cruinneachaidh
luchd-dìon
Luchd-dìon
luchd-ealain
Luchd-ealain
luchd-einnseanaraidh
Luchd-einnseanaraidh
luchd-glèidhteachais
Luchd-glèidhteachais
luchd-gnìomhachais
Luchd-gnìomhachais
luchd-iomairt
Luchd-iomairt
luchd-lagh
Luchd-lagh
luchd-lagha
Luchd-lagha
luchd-leanmhainn
Luchd-leanmhainn
luchd-litreachais
Luchd-litreachais
luchd-obrach
Luchd-obrach
luchd-reic
Luchd-reic
luchd-sgrùdaidh
Luchd-sgrùdaidh
luchd-teagaisg
Luchd-teagaisg
luchd-turais
Luchd-turais
luchd-éisdeachd
Luchd-éisdeachd
luchd-éisteachd
Luchd-éisteachd
là-an-diugh
Là-an-diugh
làmh-chuideachaidh
Làmh-chuideachaidh
làmh-sgrìobhainn
Làmh-sgrìobhainn
làmh-sgrìobhainnean
Làmh-sgrìobhainnean
làmh-sgrìobhta
Làmh-sgrìobhta
làn-bheachd
Làn-bheachd
làn-ghàidhealtachd
Làn-ghàidhealtachd
làn-thuigse
Làn-thuigse
làn-ùine
Làn-ùine
làrna-mhàireach
Làrna-mhàireach
lìn-bheaga
Lìn-bheaga
lùth-chleasan
Lùth-chleasan
ma-ta
Ma-ta
ma-tha
Ma-tha
ma-thà
Ma-thà
ma-tà
Ma-tà
mac-an-duine
Mac-an-duine
mac-léinn
Mac-léinn
mac-meanmna
Mac-meanmna
maighstir-sgoile
Maighstir-sgoile
maor-chladaich
Maor-chladaich
maor-fearainn
Maor-fearainn
mar-thà
Mar-thà
marbh-riaghailt
Marbh-riaghailt
meadhan-aoiseil
Meadhan-aoiseil
meadhan-latha
Meadhan-latha
meadhan-oidhche
Meadhan-oidhche
meal-an-naidheachd
Meal-an-naidheachd
mean-fhàs
Mean-fhàs
mhac-meanmna
Mhac-meanmna
mheadhain-latha
Mheadhain-latha
mheadhain-oidhche
Mheadhain-oidhche
mheadhan-oidhche
Mheadhan-oidhche
mheantraiginn-sa
Mheantraiginn-sa
mhi-rùn
Mhi-rùn
mhic-an-duine
Mhic-an-duine
mhoraltachd-sa
Mhoraltachd-sa
mhuir-làn
Mhuir-làn
mhuir-sgèin
Mhuir-sgèin
mhàthair-san
Mhàthair-san
mhì-chinnt
Mhì-chinnt
mhì-chneasda
Mhì-chneasda
mhì-chòrdadh
Mhì-chòrdadh
mhì-riaraichte
Mhì-riaraichte
mhì-shocair
Mhì-shocair
mhòr-chuid
Mhòr-chuid
mhòr-shluagh
Mhòr-shluagh
mhòr-shluaigh
Mhòr-shluaigh
mhór-amharas
Mhór-amharas
mhór-chuid
Mhór-chuid
mhór-shluaigh
Mhór-shluaigh
mi-chneasda
Mi-chneasda
mi-rùn
Mi-rùn
mic-léinn
Mic-léinn
mion-chànain
Mion-chànain
mion-fhios
Mion-fhios
mion-fhiosrach
Mion-fhiosrach
mion-sgrùdadh
Mion-sgrùdadh
muir-meadhon-thireach
Muir-meadhon-thireach
mèinnean-talmhainn
Mèinnean-talmhainn
mì-chinnt
Mì-chinnt
mì-choltach
Mì-choltach
mì-dhòigh
Mì-dhòigh
mì-fhair
Mì-fhair
mì-fhortanach
Mì-fhortanach
mì-laghail
Mì-laghail
mì-nàdarra
Mì-nàdarra
mì-nàdarrach
Mì-nàdarrach
mì-rùin
Mì-rùin
mì-shealbhach
Mì-shealbhach
mì-thlachd
Mì-thlachd
mòr-shluagh
Mòr-shluagh
mór-bhuannachd
Mór-bhuannachd
mór-chuid
Mór-chuid
mór-roinn
Mór-roinn
n-
N-
neach-casaid
Neach-casaid
neach-cathrach
Neach-cathrach
neach-gairm
Neach-gairm
neo-chiontach
Neo-chiontach
neo-eisimeileach
Neo-eisimeileach
neo-iomlan
Neo-iomlan
neo-àbhaisteach
Neo-àbhaisteach
nua-bhàrdachd
Nua-bhàrdachd
nì-eigin
Nì-eigin
obair-sa
Obair-sa
oifigear-stiùiridh
Oifigear-stiùiridh
oirbh-se
Oirbh-se
ola-thruis
Ola-thruis
orm-sa
Orm-sa
orra-san
Orra-san
phiuthar-chéile
Phiuthar-chéile
phort-adhair
Phort-adhair
phump-adh
Phump-adh
phàipeir-naidheachd
Phàipeir-naidheachd
phòcaid-thòine
Phòcaid-thòine
pole-aichean
Pole-aichean
port-adhair
Port-adhair
proove-adh
Proove-adh
pàipear-naidheachd
Pàipear-naidheachd
pàipearan-naidheachd
Pàipearan-naidheachd
radio-beò
Radio-beò
rithe-se
Rithe-se
rium-sa
Rium-sa
ro-chumhang
Ro-chumhang
ro-eòlach
Ro-eòlach
ro-innleachd
Ro-innleachd
ro-làimh
Ro-làimh
ro-shealladh
Ro-shealladh
roth-thoisich
Roth-thoisich
rèidio-beò
Rèidio-beò
rùm-cùil
Rùm-cùil
sadadh-a-steach
Sadadh-a-steach
samhradh-a-chaidh
Samhradh-a-chaidh
saor-làithean
Saor-làithean
sead-fhighe
Sead-fhighe
sean-ghnàthas
Sean-ghnàthas
seana-bhliadhn'
Seana-bhliadhn'
seirbhis-aisig
Seirbhis-aisig
seòl-mara
Seòl-mara
seòmar-cadail
Seòmar-cadail
sgeulachdan-gaisge
Sgeulachdan-gaisge
sgoil-marcaidheachd
Sgoil-marcaidheachd
sgìr-easbaig
Sgìr-easbaig
sgìre-easbaig
Sgìre-easbaig
sheann-fhasanta
Sheann-fhasanta
shlatan-connaidh
Shlatan-connaidh
shon-sa
Shon-sa
shàr-sgoilear
Shàr-sgoilear
sibh-se
Sibh-se
snodha-gàire
Snodha-gàire
so-labhairt
So-labhairt
soch-mhalairteach
Soch-mhalairteach
spor-gunna
Spor-gunna
sàr-bheachdan
Sàr-bheachdan
sìor-dhol
Sìor-dhol
sùil-air-ais
Sùil-air-ais
sùil-mhara
Sùil-mhara
t-
T-
taigh-cuibhle
Taigh-cuibhle
taigh-céilidh
Taigh-céilidh
taigh-sa
Taigh-sa
taigh-sheinnse
Taigh-sheinnse
taigh-tasgaidh
Taigh-tasgaidh
taigh-tughaidh
Taigh-tughaidh
taigh-òsda
Taigh-òsda
taigh-òsta
Taigh-òsta
taighean-aoigheachd
Taighean-aoigheachd
taobh-sa
Taobh-sa
teachd-an-tìr
Teachd-an-tìr
teaghlach-chànanan
Teaghlach-chànanan
thaicean-airgid
Thaicean-airgid
thaighean-altraim
Thaighean-altraim
thonn-gheal
Thonn-gheal
thuigse-san
Thuigse-san
tigh-croiteir
Tigh-croiteir
tigh-còmhnaidh
Tigh-còmhnaidh
tigh-seinnse
Tigh-seinnse
tigh-sheinnse
Tigh-sheinnse
tighearnan-fearainn
Tighearnan-fearainn
togail-cridhe
Togail-cridhe
travel-adh
Travel-adh
triob-sa
Triob-sa
tro-chèile
Tro-chèile
troimh-a-chéile
Troimh-a-chéile
troimh-chèile
Troimh-chèile
troimhe-chéile
Troimhe-chéile
tuathanas-éisg
Tuathanas-éisg
tè-labhairt
Tè-labhairt
tìr-mhóir
Tìr-mhóir
tìr-mòr
Tìr-mòr
ugam-s'
Ugam-s'
ugam-sa
Ugam-sa
uige-san
Uige-san
uile-gu-lèir
Uile-gu-lèir
uile-tuigseach
Uile-tuigseach
use-agadh
Use-agadh
watch-adh
Watch-adh
weld-adh
Weld-adh
àrd-cheannard
Àrd-cheannard
àrd-chomhairliche
Àrd-chomhairliche
àrd-chonstabal
Àrd-chonstabal
àrd-dhuine
Àrd-dhuine
àrd-ionmhair
Àrd-ionmhair
àrd-oifigear
Àrd-oifigear
àrd-oifigeir
Àrd-oifigeir
àrd-sgoil
Àrd-sgoil
àrd-ìre
Àrd-ìre
àrd-ùrlair
Àrd-ùrlair
àrd-ùrlar
Àrd-ùrlar
às-creideach
Às-creideach
àtha-cheilpe
Àtha-cheilpe
ìre-sa
Ìre-sa
ìre-se
Ìre-se
òg-mhios
Òg-mhios
òige-sa
Òige-sa
òrd-mhòr
Òrd-mhòr""".split():
    _exc[orth] = [{ORTH: orth}]

# Multiple words that should remain as one token
for orth in """'n diugh
'N diugh
'n dà
'N dà
'n iar
'N iar
'n seo
'N seo
'n uairsin
'N uairsin
a a sineach
A a sineach
a b'
A b'
a bhos
A bhos
a bhàn
A bhàn
a bhòn raoir
A bhòn raoir
a bhòn uiridh
A bhòn uiridh
a bhòn-dè
A bhòn-dè
a bhòn-raoir
A bhòn-raoir
a bhòn-uiridh
A bhòn-uiridh
a bu'
A bu'
a chaoidh
A chaoidh
a cheana
A cheana
a chionn
A chionn
a chionn 's
A chionn 's
a chuile
A chuile
a chèil
A chèil
a chèile
A chèile
a chéile
A chéile
a deas
A deas
a dh'
A dh'
a h-uile
A h-uile
a mach
A mach
a muigh
A muigh
a màireach
A màireach
a nall
A nall
a neisd
A neisd
a nis
A nis
a nisd
A nisd
a nise
A nise
a niste
A niste
a nochd
A nochd
a nuas
A nuas
a null
A null
a raoir
A raoir
a riamh
A riamh
a rithist
A rithist
a s
A s
a seo
A seo
a seothach
A seothach
a shineach
A shineach
a sin
A sin
a sineach
A sineach
a staidh
A staidh
a staigh
A staigh
a steach
A steach
a stigh
A stigh
a tuath
A tuath
a uiridh
A uiridh
a' diugh
A' diugh
a' s
A' s
air bith
Air bith
air choireigin
Air choireigin
air choireigin-ach
Air choireigin-ach
air choreigin
Air choreigin
air dheireadh
Air dheireadh
air falbh
Air falbh
air neo
Air neo
air thùs
Air thùs
am a màireach muigh
Am a màireach muigh
am bliadhna
Am bliadhna
am muigh
Am muigh
an am
An am
an aodann bàn
An aodann bàn
an ath bhliadhna
An ath bhliadhna
an ath oidhch'
An ath oidhch'
an ath oidhche
An ath oidhche
an ath sheachdain
An ath sheachdain
an ath sheachdainn
An ath sheachdainn
an ath-bhliadhna
An ath-bhliadhna
an ath-oidhch'
An ath-oidhch'
an ath-oidhche
An ath-oidhche
an ath-sheachdain
An ath-sheachdain
an ath-sheachdainn
An ath-sheachdainn
an ceart-uair
An ceart-uair
an ceartuair
An ceartuair
an còmhnaidh
An còmhnaidh
an de
An de
an deas
An deas
an diugh
An diugh
an dràsda
An dràsda
an dràsta
An dràsta
an dè
An dè
an ear
An ear
an earair
An earair
an earar
An earar
an earras
An earras
an iar
An iar
an iaras
An iaras
an làrna-mhàireach
An làrna-mhàireach
an raoir
An raoir
an sean
An sean
an seo
An seo
an seothach
An seothach
an sin
An sin
an sineach
An sineach
an siod
An siod
an siud
An siud
an siudach
An siudach
an toiseach
An toiseach
an uair
An uair
an uair sin
An uair sin
an uairsin
An uairsin
an uirigh
An uirigh
an àird
An àird
an àirde
An àirde
an ìre mhath
An ìre mhath
ana nàdarra
Ana nàdarra
ann a
Ann a
ann a sheo
Ann a sheo
ann a sheothach
Ann a sheothach
ann a shin
Ann a shin
ann a shineach
Ann a shineach
ann a shiodach
Ann a shiodach
ann a shiud
Ann a shiud
ann a shiudach
Ann a shiudach
ann a'
Ann a'
ann a' shiudach
Ann a' shiudach
ann a-seo
Ann a-seo
ann a-seothach
Ann a-seothach
ann a-sin
Ann a-sin
ann a-sineach
Ann a-sineach
ann a-siud
Ann a-siud
ann am
Ann am
ann an
Ann an
ann an seo
Ann an seo
ann an shin
Ann an shin
ann an shiud
Ann an shiud
ann an sin
Ann an sin
ann an siud
Ann an siud
ann seo
Ann seo
anns a' bhad
Anns a' bhad
anns an
Anns an
ath-oidhch'
Ath-oidhch'
ban-righ 'nn
Ban-righ 'nn
bho thoiseach
Bho thoiseach
bhon 'n
Bhon 'n
bhon a'
Bhon a'
bhon an
Bhon an
bhrist '
Bhrist '
buille a-mach
Buille a-mach
bun os cionn
Bun os cionn
car son
Car son
ceann a tuath
Ceann a tuath
cia mheud
Cia mheud
coille chaoil
Coille chaoil
cò mheud
Cò mheud
có dhiubh
Có dhiubh
d' rachadh
D' rachadh
dhen an
Dhen an
do n
Do n
dè mar
Dè mar
dé mar
Dé mar
eilean tiridhe
Eilean tiridhe
fa leth
Fa leth
fad as
Fad as
fo dheireadh
Fo dheireadh
fon a'
Fon a'
fon an
Fon an
gar bith
Gar bith
gar bith có
Gar bith có
ge 's bith
Ge 's bith
ge b' e air bith
Ge b' e air bith
ge be
Ge be
ge brith
Ge brith
ge brì
Ge brì
gleann dail
Gleann dail
gleann ois
Gleann ois
gu bè gu dè
Gu bè gu dè
gu dè
Gu dè
gu dé
Gu dé
gu ruige
Gu ruige
ho ro gheallaidh
Ho ro gheallaidh
ma dheireadh
Ma dheireadh
ma dheireadh thall
Ma dheireadh thall
ma sgaoil
Ma sgaoil
ma tha
Ma tha
mar an ceudna
Mar an ceudna
mar bu trice
Mar bu trice
mar tha
Mar tha
meadhan aoiseil
Meadhan aoiseil
mu 'n
Mu 'n
mu chuairt
Mu chuairt
mu dheas
Mu dheas
mu dheireadh
Mu dheireadh
mu dheireadh thall
Mu dheireadh thall
mu n
Mu n
mu thràth
Mu thràth
mun a'
Mun a'
mun an
Mun an
na b'
Na b'
na bu
Na bu
na iad
Na iad
nach maireann
Nach maireann
o'n uairsin
O'n uairsin
oidhch '
Oidhch '
on a'
On a'
on an
On an
pholl a' ghrùthain
Pholl a' ghrùthain
roinn eorpa
Roinn eorpa
ron a'
Ron a'
ron an
Ron an
ruaidh mhònaidh
Ruaidh mhònaidh
ruith thairis
Ruith thairis
sa bhad
Sa bhad
sadadh a-mach
Sadadh a-mach
sadadh a-steach
Sadadh a-steach
sam bidh
Sam bidh
sam bith
Sam bith
srath chluaidh
Srath chluaidh
taobh a-muigh
Taobh a-muigh
taobh an ear
Taobh an ear
taobh an iar
Taobh an iar
tria san ngaoidhilcc nalbanaigh
Tria san ngaoidhilcc nalbanaigh
tron a'
Tron a'
tron an
Tron an
tuilleadh 's a chòir
Tuilleadh 's a chòir
tuilleadh sa chòir
Tuilleadh sa chòir""".split("\n"):
    _exc[orth] = [{ORTH: orth}]


TOKENIZER_EXCEPTIONS = update_exc(BASE_EXCEPTIONS, _exc)



# cython: infer_types=True, binding=True
import importlib
import sys
from typing import Callable, List, Optional

import srsly

from ..tokens.doc cimport Doc

from .. import util
from ..language import Language
from .pipe import Pipe
from .senter import senter_score

# see #9050
BACKWARD_OVERWRITE = False


class Sentencizer(Pipe):
    """Segment the Doc into sentences using a rule-based strategy.

    DOCS: https://spacy.io/api/sentencizer
    """

    default_punct_chars = [
        '!', '.', '?', '։', '؟', '۔', '܀', '܁', '܂', '߹',
        '।', '॥', '၊', '။', '።', '፧', '፨', '᙮', '᜵', '᜶', '᠃', '᠉', '᥄',
        '᥅', '᪨', '᪩', '᪪', '᪫', '᭚', '᭛', '᭞', '᭟', '᰻', '᰼', '᱾', '᱿',
        '‼', '‽', '⁇', '⁈', '⁉', '⸮', '⸼', '꓿', '꘎', '꘏', '꛳', '꛷', '꡶',
        '꡷', '꣎', '꣏', '꤯', '꧈', '꧉', '꩝', '꩞', '꩟', '꫰', '꫱', '꯫', '﹒',
        '﹖', '﹗', '！', '．', '？', '𐩖', '𐩗', '𑁇', '𑁈', '𑂾', '𑂿', '𑃀',
        '𑃁', '𑅁', '𑅂', '𑅃', '𑇅', '𑇆', '𑇍', '𑇞', '𑇟', '𑈸', '𑈹', '𑈻', '𑈼',
        '𑊩', '𑑋', '𑑌', '𑗂', '𑗃', '𑗉', '𑗊', '𑗋', '𑗌', '𑗍', '𑗎', '𑗏', '𑗐',
        '𑗑', '𑗒', '𑗓', '𑗔', '𑗕', '𑗖', '𑗗', '𑙁', '𑙂', '𑜼', '𑜽', '𑜾', '𑩂',
        '𑩃', '𑪛', '𑪜', '𑱁', '𑱂', '𖩮', '𖩯', '𖫵', '𖬷', '𖬸', '𖭄', '𛲟', '𝪈',
        '｡', '。'
    ]

    def __init__(
        self,
        name="sentencizer",
        *,
        punct_chars=None,
        overwrite=BACKWARD_OVERWRITE,
        scorer=senter_score,
    ):
        """Initialize the sentencizer.

        punct_chars (list): Punctuation characters to split on. Will be
            serialized with the nlp object.
        scorer (Optional[Callable]): The scoring method. Defaults to
            Scorer.score_spans for the attribute "sents".

        DOCS: https://spacy.io/api/sentencizer#init
        """
        self.name = name
        if punct_chars:
            self.punct_chars = set(punct_chars)
        else:
            self.punct_chars = set(self.default_punct_chars)
        self.overwrite = overwrite
        self.scorer = scorer

    def __call__(self, doc):
        """Apply the sentencizer to a Doc and set Token.is_sent_start.

        doc (Doc): The document to process.
        RETURNS (Doc): The processed Doc.

        DOCS: https://spacy.io/api/sentencizer#call
        """
        error_handler = self.get_error_handler()
        try:
            tags = self.predict([doc])
            self.set_annotations([doc], tags)
            return doc
        except Exception as e:
            error_handler(self.name, self, [doc], e)

    def predict(self, docs):
        """Apply the pipe to a batch of docs, without modifying them.

        docs (Iterable[Doc]): The documents to predict.
        RETURNS: The predictions for each document.
        """
        if not any(len(doc) for doc in docs):
            # Handle cases where there are no tokens in any docs.
            guesses = [[] for doc in docs]
            return guesses
        guesses = []
        for doc in docs:
            doc_guesses = [False] * len(doc)
            if len(doc) > 0:
                start = 0
                seen_period = False
                doc_guesses[0] = True
                for i, token in enumerate(doc):
                    is_in_punct_chars = token.text in self.punct_chars
                    if seen_period and not token.is_punct and not is_in_punct_chars:
                        doc_guesses[start] = True
                        start = token.i
                        seen_period = False
                    elif is_in_punct_chars:
                        seen_period = True
                if start < len(doc):
                    doc_guesses[start] = True
            guesses.append(doc_guesses)
        return guesses

    def set_annotations(self, docs, batch_tag_ids):
        """Modify a batch of documents, using pre-computed scores.

        docs (Iterable[Doc]): The documents to modify.
        scores: The tag IDs produced by Sentencizer.predict.
        """
        if isinstance(docs, Doc):
            docs = [docs]
        cdef Doc doc
        for i, doc in enumerate(docs):
            doc_tag_ids = batch_tag_ids[i]
            for j, tag_id in enumerate(doc_tag_ids):
                if doc.c[j].sent_start == 0 or self.overwrite:
                    if tag_id:
                        doc.c[j].sent_start = 1
                    else:
                        doc.c[j].sent_start = -1

    def to_bytes(self, *, exclude=tuple()):
        """Serialize the sentencizer to a bytestring.

        RETURNS (bytes): The serialized object.

        DOCS: https://spacy.io/api/sentencizer#to_bytes
        """
        return srsly.msgpack_dumps({"punct_chars": list(self.punct_chars), "overwrite": self.overwrite})

    def from_bytes(self, bytes_data, *, exclude=tuple()):
        """Load the sentencizer from a bytestring.

        bytes_data (bytes): The data to load.
        returns (Sentencizer): The loaded object.

        DOCS: https://spacy.io/api/sentencizer#from_bytes
        """
        cfg = srsly.msgpack_loads(bytes_data)
        self.punct_chars = set(cfg.get("punct_chars", self.default_punct_chars))
        self.overwrite = cfg.get("overwrite", self.overwrite)
        return self

    def to_disk(self, path, *, exclude=tuple()):
        """Serialize the sentencizer to disk.

        DOCS: https://spacy.io/api/sentencizer#to_disk
        """
        path = util.ensure_path(path)
        path = path.with_suffix(".json")
        srsly.write_json(path, {"punct_chars": list(self.punct_chars), "overwrite": self.overwrite})

    def from_disk(self, path, *, exclude=tuple()):
        """Load the sentencizer from disk.

        DOCS: https://spacy.io/api/sentencizer#from_disk
        """
        path = util.ensure_path(path)
        path = path.with_suffix(".json")
        cfg = srsly.read_json(path)
        self.punct_chars = set(cfg.get("punct_chars", self.default_punct_chars))
        self.overwrite = cfg.get("overwrite", self.overwrite)
        return self


# Setup backwards compatibility hook for factories
def __getattr__(name):
    if name == "make_sentencizer":
        module = importlib.import_module("spacy.pipeline.factories")
        return module.make_sentencizer
    raise AttributeError(f"module {__name__} has no attribute {name}")



"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.kmr.examples import sentences
>>> docs = nlp.pipe(sentences)
"""

sentences = [
    "Berê mirovan her tim li geşedana pêşerojê ye",  # People's gaze is always on the development of the future
    "Kawa Nemir di 14 salan de Ulysses wergerand Kurmancî.",  # Kawa Nemir translated Ulysses into Kurmanji in 14 years.
    "Mem Ararat hunermendekî Kurd yê bi nav û deng e.",  # Mem Ararat is a famous Kurdish artist
    "Firat Cewerî 40 sal e pirtûkên Kurdî dinivîsîne.",  # Firat Ceweri has been writing Kurdish books for 40 years
    "Rojnamegerê ciwan nûçeyeke balkêş li ser rewşa aborî nivîsand",  # The young journalist wrote an interesting news article about the economic situation
    "Sektora çandiniyê beşeke giring a belavkirina gaza serayê li seranserê cîhanê pêk tîne",  # The agricultural sector constitutes an important part of greenhouse gas emissions worldwide
    "Xwendekarên jêhatî di pêşbaziya matematîkê de serkeftî bûn",  # Talented students succeeded in the mathematics competition
    "Ji ber ji tunebûnê bavê min xwişkeke min nedan xwendin ew ji min re bû derd û kulek.",  # Because of poverty, my father didn't send my sister to school, which became a pain and sorrow for me
]



STOP_WORDS = set(
    """
'ad
'ar
'd # iad
'g # ag
'ga
'gam
'gan
'gar
'gur
'm # am
'n # an
'n seo
'na
'nad
'nam
'nan
'nar
'nuair
'nur
's
'sa
'san
'sann
'se
'sna
a
a'
a'd # agad
a'm # agam
a-chèile
a-seo
a-sin
a-siud
a chionn
a chionn 's
a chèile
a chéile
a dh'
a h-uile
a seo
ac' # aca
aca
aca-san
acasan
ach
ag
agad
agad-sa
agads'
agadsa
agaibh
agaibhse
againn
againne
agam
agam-sa
agams'
agamsa
agus
aice
aice-se
aicese
aig
aig' # aige
aige
aige-san
aigesan
air
air-san
air neo
airsan
am
an
an seo
an sin
an siud
an uair
ann
ann a
ann a'
ann a shin
ann am
ann an
annad
annam
annam-s'
annamsa
anns
anns an
annta
aon
ar
as
asad
asda
asta
b'
bho
bhon
bhuaidhe # bhuaithe
bhuainn
bhuaipe
bhuaithe
bhuapa
bhur
brì
bu
c'à
car son
carson
cha
chan
chionn
choir
chon
chun
chèile
chéile
chòir
cia mheud
ciamar
co-dhiubh
cuide
cuin
cuin'
cuine
cà
cà'
càil
càit
càit'
càite
cò
cò mheud
có
d'
da
de
dh'
dha
dhaibh
dhaibh-san
dhaibhsan
dhan
dhasan
dhe
dhen
dheth
dhi
dhiom
dhiot
dhith
dhiubh
dhomh
dhomh-s'
dhomhsa
dhu'sa # dhut-sa
dhuibh
dhuibhse
dhuinn
dhuinne
dhuit
dhut
dhutsa
dhut-sa
dhà
dhà-san
dhàsan
dhòmhsa
diubh
do
docha
don
dà
dè
dè mar
dé
dé mar
dòch'
dòcha
e
eadar
eatarra
eatorra
eile
esan
fa
far
feud
fhad
fheudar
fhearr
fhein
fheudar
fheàrr
fhèin
fhéin
fhìn
fo
fodha
fodhainn
foipe
fon
fèin
ga
gach
gam
gan
ge brith
ged
gu
gu dè
gu ruige
gun
gur
gus
i
iad
iadsan
innte
is
ise
le
leam
leam-sa
leamsa
leat
leat-sa
leatha
leatsa
leibh
leis
leis-san
leoth'
leotha
leotha-san
linn
m'
m'a
ma
mac
man
mar
mas
mathaid
mi
mis'
mise
mo
mu
mu 'n
mun
mur
mura
mus
na
na b'
na bu
na iad
nach
nad
nam
nan
nar
nas
neo
no
nuair
o
o'n
oir
oirbh
oirbh-se
oirnn
oirnne
oirre
on
orm
orm-sa
ormsa
orra
orra-san
orrasan
ort
os
r'
ri
ribh
rinn
ris
rithe
rithe-se
rium
rium-sa
riums'
riumsa
riut
riuth'
riutha
riuthasan
ro
ro'n
roimh
roimhe
romhainn
romham
romhpa
ron
ruibh
ruinn
ruinne
sa
san
sann
se
seach
seo
seothach
shin
sibh
sibh-se
sibhse
sin
sineach
sinn
sinne
siod
siodach
siud
siudach
sna # ann an
sè
t'
tarsaing
tarsainn
tarsuinn
thar
thoigh
thro
thu
thuc'
thuca
thugad
thugaibh
thugainn
thugam
thugamsa
thuice
thuige
thus'
thusa
timcheall
toigh
toil
tro
tro' # troimh
troimh
troimhe
tron
tu
tusa
uair
ud
ugaibh
ugam-s'
ugam-sa
uice
uige
uige-san
umad
unnta # ann an
ur
urrainn
à
às
àsan
á
ás
è
ì
ò
ó
""".split("\n")
)



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from thinc.api import Model
from thinc.types import Floats2d, Ragged

from ..kb import Candidate, KnowledgeBase
from ..language import Language
from ..pipeline._parser_internals.transition_system import TransitionSystem
from ..pipeline.attributeruler import AttributeRuler
from ..pipeline.dep_parser import DEFAULT_PARSER_MODEL, DependencyParser
from ..pipeline.edit_tree_lemmatizer import (
    DEFAULT_EDIT_TREE_LEMMATIZER_MODEL,
    EditTreeLemmatizer,
)

# Import factory default configurations
from ..pipeline.entity_linker import DEFAULT_NEL_MODEL, EntityLinker
from ..pipeline.entityruler import DEFAULT_ENT_ID_SEP, EntityRuler
from ..pipeline.functions import DocCleaner, TokenSplitter
from ..pipeline.legacy import EntityLinker_v1
from ..pipeline.lemmatizer import Lemmatizer
from ..pipeline.morphologizer import DEFAULT_MORPH_MODEL, Morphologizer
from ..pipeline.multitask import DEFAULT_MT_MODEL, MultitaskObjective
from ..pipeline.ner import DEFAULT_NER_MODEL, EntityRecognizer
from ..pipeline.sentencizer import Sentencizer
from ..pipeline.senter import DEFAULT_SENTER_MODEL, SentenceRecognizer
from ..pipeline.span_finder import DEFAULT_SPAN_FINDER_MODEL, SpanFinder
from ..pipeline.span_ruler import (
    DEFAULT_SPANS_KEY as SPAN_RULER_DEFAULT_SPANS_KEY,
    SpanRuler,
    prioritize_existing_ents_filter,
    prioritize_new_ents_filter,
)
from ..pipeline.spancat import (
    DEFAULT_SPANCAT_MODEL,
    DEFAULT_SPANCAT_SINGLELABEL_MODEL,
    DEFAULT_SPANS_KEY,
    SpanCategorizer,
    Suggester,
)
from ..pipeline.tagger import DEFAULT_TAGGER_MODEL, Tagger
from ..pipeline.textcat import DEFAULT_SINGLE_TEXTCAT_MODEL, TextCategorizer
from ..pipeline.textcat_multilabel import (
    DEFAULT_MULTI_TEXTCAT_MODEL,
    MultiLabel_TextCategorizer,
)
from ..pipeline.tok2vec import DEFAULT_TOK2VEC_MODEL, Tok2Vec
from ..tokens.doc import Doc
from ..tokens.span import Span
from ..vocab import Vocab

# Global flag to track if factories have been registered
FACTORIES_REGISTERED = False


def register_factories() -> None:
    """Register all factories with the registry.

    This function registers all pipeline component factories, centralizing
    the registrations that were previously done with @Language.factory decorators.
    """
    global FACTORIES_REGISTERED

    if FACTORIES_REGISTERED:
        return

    # Register factories using the same pattern as Language.factory decorator
    # We use Language.factory()() pattern which exactly mimics the decorator

    # attributeruler
    Language.factory(
        "attribute_ruler",
        default_config={
            "validate": False,
            "scorer": {"@scorers": "spacy.attribute_ruler_scorer.v1"},
        },
    )(make_attribute_ruler)

    # entity_linker
    Language.factory(
        "entity_linker",
        requires=["doc.ents", "doc.sents", "token.ent_iob", "token.ent_type"],
        assigns=["token.ent_kb_id"],
        default_config={
            "model": DEFAULT_NEL_MODEL,
            "labels_discard": [],
            "n_sents": 0,
            "incl_prior": True,
            "incl_context": True,
            "entity_vector_length": 64,
            "get_candidates": {"@misc": "spacy.CandidateGenerator.v1"},
            "get_candidates_batch": {"@misc": "spacy.CandidateBatchGenerator.v1"},
            "generate_empty_kb": {"@misc": "spacy.EmptyKB.v2"},
            "overwrite": True,
            "scorer": {"@scorers": "spacy.entity_linker_scorer.v1"},
            "use_gold_ents": True,
            "candidates_batch_size": 1,
            "threshold": None,
        },
        default_score_weights={
            "nel_micro_f": 1.0,
            "nel_micro_r": None,
            "nel_micro_p": None,
        },
    )(make_entity_linker)

    # entity_ruler
    Language.factory(
        "entity_ruler",
        assigns=["doc.ents", "token.ent_type", "token.ent_iob"],
        default_config={
            "phrase_matcher_attr": None,
            "matcher_fuzzy_compare": {"@misc": "spacy.levenshtein_compare.v1"},
            "validate": False,
            "overwrite_ents": False,
            "ent_id_sep": DEFAULT_ENT_ID_SEP,
            "scorer": {"@scorers": "spacy.entity_ruler_scorer.v1"},
        },
        default_score_weights={
            "ents_f": 1.0,
            "ents_p": 0.0,
            "ents_r": 0.0,
            "ents_per_type": None,
        },
    )(make_entity_ruler)

    # lemmatizer
    Language.factory(
        "lemmatizer",
        assigns=["token.lemma"],
        default_config={
            "model": None,
            "mode": "lookup",
            "overwrite": False,
            "scorer": {"@scorers": "spacy.lemmatizer_scorer.v1"},
        },
        default_score_weights={"lemma_acc": 1.0},
    )(make_lemmatizer)

    # textcat
    Language.factory(
        "textcat",
        assigns=["doc.cats"],
        default_config={
            "threshold": 0.0,
            "model": DEFAULT_SINGLE_TEXTCAT_MODEL,
            "scorer": {"@scorers": "spacy.textcat_scorer.v2"},
        },
        default_score_weights={
            "cats_score": 1.0,
            "cats_score_desc": None,
            "cats_micro_p": None,
            "cats_micro_r": None,
            "cats_micro_f": None,
            "cats_macro_p": None,
            "cats_macro_r": None,
            "cats_macro_f": None,
            "cats_macro_auc": None,
            "cats_f_per_type": None,
        },
    )(make_textcat)

    # token_splitter
    Language.factory(
        "token_splitter",
        default_config={"min_length": 25, "split_length": 10},
        retokenizes=True,
    )(make_token_splitter)

    # doc_cleaner
    Language.factory(
        "doc_cleaner",
        default_config={"attrs": {"tensor": None, "_.trf_data": None}, "silent": True},
    )(make_doc_cleaner)

    # tok2vec
    Language.factory(
        "tok2vec",
        assigns=["doc.tensor"],
        default_config={"model": DEFAULT_TOK2VEC_MODEL},
    )(make_tok2vec)

    # senter
    Language.factory(
        "senter",
        assigns=["token.is_sent_start"],
        default_config={
            "model": DEFAULT_SENTER_MODEL,
            "overwrite": False,
            "scorer": {"@scorers": "spacy.senter_scorer.v1"},
        },
        default_score_weights={"sents_f": 1.0, "sents_p": 0.0, "sents_r": 0.0},
    )(make_senter)

    # morphologizer
    Language.factory(
        "morphologizer",
        assigns=["token.morph", "token.pos"],
        default_config={
            "model": DEFAULT_MORPH_MODEL,
            "overwrite": True,
            "extend": False,
            "scorer": {"@scorers": "spacy.morphologizer_scorer.v1"},
            "label_smoothing": 0.0,
        },
        default_score_weights={
            "pos_acc": 0.5,
            "morph_acc": 0.5,
            "morph_per_feat": None,
        },
    )(make_morphologizer)

    # spancat
    Language.factory(
        "spancat",
        assigns=["doc.spans"],
        default_config={
            "threshold": 0.5,
            "spans_key": DEFAULT_SPANS_KEY,
            "max_positive": None,
            "model": DEFAULT_SPANCAT_MODEL,
            "suggester": {"@misc": "spacy.ngram_suggester.v1", "sizes": [1, 2, 3]},
            "scorer": {"@scorers": "spacy.spancat_scorer.v1"},
        },
        default_score_weights={"spans_sc_f": 1.0, "spans_sc_p": 0.0, "spans_sc_r": 0.0},
    )(make_spancat)

    # spancat_singlelabel
    Language.factory(
        "spancat_singlelabel",
        assigns=["doc.spans"],
        default_config={
            "spans_key": DEFAULT_SPANS_KEY,
            "model": DEFAULT_SPANCAT_SINGLELABEL_MODEL,
            "negative_weight": 1.0,
            "suggester": {"@misc": "spacy.ngram_suggester.v1", "sizes": [1, 2, 3]},
            "scorer": {"@scorers": "spacy.spancat_scorer.v1"},
            "allow_overlap": True,
        },
        default_score_weights={"spans_sc_f": 1.0, "spans_sc_p": 0.0, "spans_sc_r": 0.0},
    )(make_spancat_singlelabel)

    # future_entity_ruler
    Language.factory(
        "future_entity_ruler",
        assigns=["doc.ents"],
        default_config={
            "phrase_matcher_attr": None,
            "validate": False,
            "overwrite_ents": False,
            "scorer": {"@scorers": "spacy.entity_ruler_scorer.v1"},
            "ent_id_sep": "__unused__",
            "matcher_fuzzy_compare": {"@misc": "spacy.levenshtein_compare.v1"},
        },
        default_score_weights={
            "ents_f": 1.0,
            "ents_p": 0.0,
            "ents_r": 0.0,
            "ents_per_type": None,
        },
    )(make_future_entity_ruler)

    # span_ruler
    Language.factory(
        "span_ruler",
        assigns=["doc.spans"],
        default_config={
            "spans_key": SPAN_RULER_DEFAULT_SPANS_KEY,
            "spans_filter": None,
            "annotate_ents": False,
            "ents_filter": {"@misc": "spacy.first_longest_spans_filter.v1"},
            "phrase_matcher_attr": None,
            "matcher_fuzzy_compare": {"@misc": "spacy.levenshtein_compare.v1"},
            "validate": False,
            "overwrite": True,
            "scorer": {
                "@scorers": "spacy.overlapping_labeled_spans_scorer.v1",
                "spans_key": SPAN_RULER_DEFAULT_SPANS_KEY,
            },
        },
        default_score_weights={
            f"spans_{SPAN_RULER_DEFAULT_SPANS_KEY}_f": 1.0,
            f"spans_{SPAN_RULER_DEFAULT_SPANS_KEY}_p": 0.0,
            f"spans_{SPAN_RULER_DEFAULT_SPANS_KEY}_r": 0.0,
            f"spans_{SPAN_RULER_DEFAULT_SPANS_KEY}_per_type": None,
        },
    )(make_span_ruler)

    # trainable_lemmatizer
    Language.factory(
        "trainable_lemmatizer",
        assigns=["token.lemma"],
        requires=[],
        default_config={
            "model": DEFAULT_EDIT_TREE_LEMMATIZER_MODEL,
            "backoff": "orth",
            "min_tree_freq": 3,
            "overwrite": False,
            "top_k": 1,
            "scorer": {"@scorers": "spacy.lemmatizer_scorer.v1"},
        },
        default_score_weights={"lemma_acc": 1.0},
    )(make_edit_tree_lemmatizer)

    # textcat_multilabel
    Language.factory(
        "textcat_multilabel",
        assigns=["doc.cats"],
        default_config={
            "threshold": 0.5,
            "model": DEFAULT_MULTI_TEXTCAT_MODEL,
            "scorer": {"@scorers": "spacy.textcat_multilabel_scorer.v2"},
        },
        default_score_weights={
            "cats_score": 1.0,
            "cats_score_desc": None,
            "cats_micro_p": None,
            "cats_micro_r": None,
            "cats_micro_f": None,
            "cats_macro_p": None,
            "cats_macro_r": None,
            "cats_macro_f": None,
            "cats_macro_auc": None,
            "cats_f_per_type": None,
        },
    )(make_multilabel_textcat)

    # span_finder
    Language.factory(
        "span_finder",
        assigns=["doc.spans"],
        default_config={
            "threshold": 0.5,
            "model": DEFAULT_SPAN_FINDER_MODEL,
            "spans_key": DEFAULT_SPANS_KEY,
            "max_length": 25,
            "min_length": None,
            "scorer": {"@scorers": "spacy.span_finder_scorer.v1"},
        },
        default_score_weights={
            f"spans_{DEFAULT_SPANS_KEY}_f": 1.0,
            f"spans_{DEFAULT_SPANS_KEY}_p": 0.0,
            f"spans_{DEFAULT_SPANS_KEY}_r": 0.0,
        },
    )(make_span_finder)

    # ner
    Language.factory(
        "ner",
        assigns=["doc.ents", "token.ent_iob", "token.ent_type"],
        default_config={
            "moves": None,
            "update_with_oracle_cut_size": 100,
            "model": DEFAULT_NER_MODEL,
            "incorrect_spans_key": None,
            "scorer": {"@scorers": "spacy.ner_scorer.v1"},
        },
        default_score_weights={
            "ents_f": 1.0,
            "ents_p": 0.0,
            "ents_r": 0.0,
            "ents_per_type": None,
        },
    )(make_ner)

    # beam_ner
    Language.factory(
        "beam_ner",
        assigns=["doc.ents", "token.ent_iob", "token.ent_type"],
        default_config={
            "moves": None,
            "update_with_oracle_cut_size": 100,
            "model": DEFAULT_NER_MODEL,
            "beam_density": 0.01,
            "beam_update_prob": 0.5,
            "beam_width": 32,
            "incorrect_spans_key": None,
            "scorer": {"@scorers": "spacy.ner_scorer.v1"},
        },
        default_score_weights={
            "ents_f": 1.0,
            "ents_p": 0.0,
            "ents_r": 0.0,
            "ents_per_type": None,
        },
    )(make_beam_ner)

    # parser
    Language.factory(
        "parser",
        assigns=["token.dep", "token.head", "token.is_sent_start", "doc.sents"],
        default_config={
            "moves": None,
            "update_with_oracle_cut_size": 100,
            "learn_tokens": False,
            "min_action_freq": 30,
            "model": DEFAULT_PARSER_MODEL,
            "scorer": {"@scorers": "spacy.parser_scorer.v1"},
        },
        default_score_weights={
            "dep_uas": 0.5,
            "dep_las": 0.5,
            "dep_las_per_type": None,
            "sents_p": None,
            "sents_r": None,
            "sents_f": 0.0,
        },
    )(make_parser)

    # beam_parser
    Language.factory(
        "beam_parser",
        assigns=["token.dep", "token.head", "token.is_sent_start", "doc.sents"],
        default_config={
            "moves": None,
            "update_with_oracle_cut_size": 100,
            "learn_tokens": False,
            "min_action_freq": 30,
            "beam_width": 8,
            "beam_density": 0.0001,
            "beam_update_prob": 0.5,
            "model": DEFAULT_PARSER_MODEL,
            "scorer": {"@scorers": "spacy.parser_scorer.v1"},
        },
        default_score_weights={
            "dep_uas": 0.5,
            "dep_las": 0.5,
            "dep_las_per_type": None,
            "sents_p": None,
            "sents_r": None,
            "sents_f": 0.0,
        },
    )(make_beam_parser)

    # tagger
    Language.factory(
        "tagger",
        assigns=["token.tag"],
        default_config={
            "model": DEFAULT_TAGGER_MODEL,
            "overwrite": False,
            "scorer": {"@scorers": "spacy.tagger_scorer.v1"},
            "neg_prefix": "!",
            "label_smoothing": 0.0,
        },
        default_score_weights={
            "tag_acc": 1.0,
            "pos_acc": 0.0,
            "tag_micro_p": None,
            "tag_micro_r": None,
            "tag_micro_f": None,
        },
    )(make_tagger)

    # nn_labeller
    Language.factory(
        "nn_labeller",
        default_config={
            "labels": None,
            "target": "dep_tag_offset",
            "model": DEFAULT_MT_MODEL,
        },
    )(make_nn_labeller)

    # sentencizer
    Language.factory(
        "sentencizer",
        assigns=["token.is_sent_start", "doc.sents"],
        default_config={
            "punct_chars": None,
            "overwrite": False,
            "scorer": {"@scorers": "spacy.senter_scorer.v1"},
        },
        default_score_weights={"sents_f": 1.0, "sents_p": 0.0, "sents_r": 0.0},
    )(make_sentencizer)

    # Set the flag to indicate that all factories have been registered
    FACTORIES_REGISTERED = True


# We can't have function implementations for these factories in Cython, because
# we need to build a Pydantic model for them dynamically, reading their argument
# structure from the signature. In Cython 3, this doesn't work because the
# from __future__ import annotations semantics are used, which means the types
# are stored as strings.
def make_sentencizer(
    nlp: Language,
    name: str,
    punct_chars: Optional[List[str]],
    overwrite: bool,
    scorer: Optional[Callable],
):
    return Sentencizer(
        name, punct_chars=punct_chars, overwrite=overwrite, scorer=scorer
    )


def make_attribute_ruler(
    nlp: Language, name: str, validate: bool, scorer: Optional[Callable]
):
    return AttributeRuler(nlp.vocab, name, validate=validate, scorer=scorer)


def make_entity_linker(
    nlp: Language,
    name: str,
    model: Model,
    *,
    labels_discard: Iterable[str],
    n_sents: int,
    incl_prior: bool,
    incl_context: bool,
    entity_vector_length: int,
    get_candidates: Callable[[KnowledgeBase, Span], Iterable[Candidate]],
    get_candidates_batch: Callable[
        [KnowledgeBase, Iterable[Span]], Iterable[Iterable[Candidate]]
    ],
    generate_empty_kb: Callable[[Vocab, int], KnowledgeBase],
    overwrite: bool,
    scorer: Optional[Callable],
    use_gold_ents: bool,
    candidates_batch_size: int,
    threshold: Optional[float] = None,
):

    if not model.attrs.get("include_span_maker", False):
        # The only difference in arguments here is that use_gold_ents and threshold aren't available.
        return EntityLinker_v1(
            nlp.vocab,
            model,
            name,
            labels_discard=labels_discard,
            n_sents=n_sents,
            incl_prior=incl_prior,
            incl_context=incl_context,
            entity_vector_length=entity_vector_length,
            get_candidates=get_candidates,
            overwrite=overwrite,
            scorer=scorer,
        )
    return EntityLinker(
        nlp.vocab,
        model,
        name,
        labels_discard=labels_discard,
        n_sents=n_sents,
        incl_prior=incl_prior,
        incl_context=incl_context,
        entity_vector_length=entity_vector_length,
        get_candidates=get_candidates,
        get_candidates_batch=get_candidates_batch,
        generate_empty_kb=generate_empty_kb,
        overwrite=overwrite,
        scorer=scorer,
        use_gold_ents=use_gold_ents,
        candidates_batch_size=candidates_batch_size,
        threshold=threshold,
    )


def make_lemmatizer(
    nlp: Language,
    model: Optional[Model],
    name: str,
    mode: str,
    overwrite: bool,
    scorer: Optional[Callable],
):
    return Lemmatizer(
        nlp.vocab, model, name, mode=mode, overwrite=overwrite, scorer=scorer
    )


def make_textcat(
    nlp: Language,
    name: str,
    model: Model[List[Doc], List[Floats2d]],
    threshold: float,
    scorer: Optional[Callable],
) -> TextCategorizer:
    return TextCategorizer(nlp.vocab, model, name, threshold=threshold, scorer=scorer)


def make_token_splitter(
    nlp: Language, name: str, *, min_length: int = 0, split_length: int = 0
):
    return TokenSplitter(min_length=min_length, split_length=split_length)


def make_doc_cleaner(nlp: Language, name: str, *, attrs: Dict[str, Any], silent: bool):
    return DocCleaner(attrs, silent=silent)


def make_tok2vec(nlp: Language, name: str, model: Model) -> Tok2Vec:
    return Tok2Vec(nlp.vocab, model, name)


def make_spancat(
    nlp: Language,
    name: str,
    suggester: Suggester,
    model: Model[Tuple[List[Doc], Ragged], Floats2d],
    spans_key: str,
    scorer: Optional[Callable],
    threshold: float,
    max_positive: Optional[int],
) -> SpanCategorizer:
    return SpanCategorizer(
        nlp.vocab,
        model=model,
        suggester=suggester,
        name=name,
        spans_key=spans_key,
        negative_weight=None,
        allow_overlap=True,
        max_positive=max_positive,
        threshold=threshold,
        scorer=scorer,
        add_negative_label=False,
    )


def make_spancat_singlelabel(
    nlp: Language,
    name: str,
    suggester: Suggester,
    model: Model[Tuple[List[Doc], Ragged], Floats2d],
    spans_key: str,
    negative_weight: float,
    allow_overlap: bool,
    scorer: Optional[Callable],
) -> SpanCategorizer:
    return SpanCategorizer(
        nlp.vocab,
        model=model,
        suggester=suggester,
        name=name,
        spans_key=spans_key,
        negative_weight=negative_weight,
        allow_overlap=allow_overlap,
        max_positive=1,
        add_negative_label=True,
        threshold=None,
        scorer=scorer,
    )


def make_future_entity_ruler(
    nlp: Language,
    name: str,
    phrase_matcher_attr: Optional[Union[int, str]],
    matcher_fuzzy_compare: Callable,
    validate: bool,
    overwrite_ents: bool,
    scorer: Optional[Callable],
    ent_id_sep: str,
):
    if overwrite_ents:
        ents_filter = prioritize_new_ents_filter
    else:
        ents_filter = prioritize_existing_ents_filter
    return SpanRuler(
        nlp,
        name,
        spans_key=None,
        spans_filter=None,
        annotate_ents=True,
        ents_filter=ents_filter,
        phrase_matcher_attr=phrase_matcher_attr,
        matcher_fuzzy_compare=matcher_fuzzy_compare,
        validate=validate,
        overwrite=False,
        scorer=scorer,
    )


def make_entity_ruler(
    nlp: Language,
    name: str,
    phrase_matcher_attr: Optional[Union[int, str]],
    matcher_fuzzy_compare: Callable,
    validate: bool,
    overwrite_ents: bool,
    ent_id_sep: str,
    scorer: Optional[Callable],
):
    return EntityRuler(
        nlp,
        name,
        phrase_matcher_attr=phrase_matcher_attr,
        matcher_fuzzy_compare=matcher_fuzzy_compare,
        validate=validate,
        overwrite_ents=overwrite_ents,
        ent_id_sep=ent_id_sep,
        scorer=scorer,
    )


def make_span_ruler(
    nlp: Language,
    name: str,
    spans_key: Optional[str],
    spans_filter: Optional[Callable[[Iterable[Span], Iterable[Span]], Iterable[Span]]],
    annotate_ents: bool,
    ents_filter: Callable[[Iterable[Span], Iterable[Span]], Iterable[Span]],
    phrase_matcher_attr: Optional[Union[int, str]],
    matcher_fuzzy_compare: Callable,
    validate: bool,
    overwrite: bool,
    scorer: Optional[Callable],
):
    return SpanRuler(
        nlp,
        name,
        spans_key=spans_key,
        spans_filter=spans_filter,
        annotate_ents=annotate_ents,
        ents_filter=ents_filter,
        phrase_matcher_attr=phrase_matcher_attr,
        matcher_fuzzy_compare=matcher_fuzzy_compare,
        validate=validate,
        overwrite=overwrite,
        scorer=scorer,
    )


def make_edit_tree_lemmatizer(
    nlp: Language,
    name: str,
    model: Model,
    backoff: Optional[str],
    min_tree_freq: int,
    overwrite: bool,
    top_k: int,
    scorer: Optional[Callable],
):
    return EditTreeLemmatizer(
        nlp.vocab,
        model,
        name,
        backoff=backoff,
        min_tree_freq=min_tree_freq,
        overwrite=overwrite,
        top_k=top_k,
        scorer=scorer,
    )


def make_multilabel_textcat(
    nlp: Language,
    name: str,
    model: Model[List[Doc], List[Floats2d]],
    threshold: float,
    scorer: Optional[Callable],
) -> MultiLabel_TextCategorizer:
    return MultiLabel_TextCategorizer(
        nlp.vocab, model, name, threshold=threshold, scorer=scorer
    )


def make_span_finder(
    nlp: Language,
    name: str,
    model: Model[Iterable[Doc], Floats2d],
    spans_key: str,
    threshold: float,
    max_length: Optional[int],
    min_length: Optional[int],
    scorer: Optional[Callable],
) -> SpanFinder:
    return SpanFinder(
        nlp,
        model=model,
        threshold=threshold,
        name=name,
        scorer=scorer,
        max_length=max_length,
        min_length=min_length,
        spans_key=spans_key,
    )


def make_ner(
    nlp: Language,
    name: str,
    model: Model,
    moves: Optional[TransitionSystem],
    update_with_oracle_cut_size: int,
    incorrect_spans_key: Optional[str],
    scorer: Optional[Callable],
):
    return EntityRecognizer(
        nlp.vocab,
        model,
        name=name,
        moves=moves,
        update_with_oracle_cut_size=update_with_oracle_cut_size,
        incorrect_spans_key=incorrect_spans_key,
        scorer=scorer,
    )


def make_beam_ner(
    nlp: Language,
    name: str,
    model: Model,
    moves: Optional[TransitionSystem],
    update_with_oracle_cut_size: int,
    beam_width: int,
    beam_density: float,
    beam_update_prob: float,
    incorrect_spans_key: Optional[str],
    scorer: Optional[Callable],
):
    return EntityRecognizer(
        nlp.vocab,
        model,
        name=name,
        moves=moves,
        update_with_oracle_cut_size=update_with_oracle_cut_size,
        beam_width=beam_width,
        beam_density=beam_density,
        beam_update_prob=beam_update_prob,
        incorrect_spans_key=incorrect_spans_key,
        scorer=scorer,
    )


def make_parser(
    nlp: Language,
    name: str,
    model: Model,
    moves: Optional[TransitionSystem],
    update_with_oracle_cut_size: int,
    learn_tokens: bool,
    min_action_freq: int,
    scorer: Optional[Callable],
):
    return DependencyParser(
        nlp.vocab,
        model,
        name=name,
        moves=moves,
        update_with_oracle_cut_size=update_with_oracle_cut_size,
        learn_tokens=learn_tokens,
        min_action_freq=min_action_freq,
        scorer=scorer,
    )


def make_beam_parser(
    nlp: Language,
    name: str,
    model: Model,
    moves: Optional[TransitionSystem],
    update_with_oracle_cut_size: int,
    learn_tokens: bool,
    min_action_freq: int,
    beam_width: int,
    beam_density: float,
    beam_update_prob: float,
    scorer: Optional[Callable],
):
    return DependencyParser(
        nlp.vocab,
        model,
        name=name,
        moves=moves,
        update_with_oracle_cut_size=update_with_oracle_cut_size,
        learn_tokens=learn_tokens,
        min_action_freq=min_action_freq,
        beam_width=beam_width,
        beam_density=beam_density,
        beam_update_prob=beam_update_prob,
        scorer=scorer,
    )


def make_tagger(
    nlp: Language,
    name: str,
    model: Model,
    overwrite: bool,
    scorer: Optional[Callable],
    neg_prefix: str,
    label_smoothing: float,
):
    return Tagger(
        nlp.vocab,
        model,
        name=name,
        overwrite=overwrite,
        scorer=scorer,
        neg_prefix=neg_prefix,
        label_smoothing=label_smoothing,
    )


def make_nn_labeller(
    nlp: Language, name: str, model: Model, labels: Optional[dict], target: str
):
    return MultitaskObjective(nlp.vocab, model, name, target=target)


def make_morphologizer(
    nlp: Language,
    model: Model,
    name: str,
    overwrite: bool,
    extend: bool,
    label_smoothing: float,
    scorer: Optional[Callable],
):
    return Morphologizer(
        nlp.vocab,
        model,
        name,
        overwrite=overwrite,
        extend=extend,
        label_smoothing=label_smoothing,
        scorer=scorer,
    )


def make_senter(
    nlp: Language, name: str, model: Model, overwrite: bool, scorer: Optional[Callable]
):
    return SentenceRecognizer(
        nlp.vocab, model, name, overwrite=overwrite, scorer=scorer
    )



# Source: https://zenodo.org/records/10148636

STOP_WORDS = set(
    """
འི་
།
དུ་
གིས་
སོགས་
ཏེ
གི་
རྣམས་
ནི
ཀུན་
ཡི་
འདི
ཀྱི་
སྙེད་
པས་
གཞན་
ཀྱིས་
ཡི
ལ
ནི་
དང་
སོགས
ཅིང་
ར
དུ
མི་
སུ་
བཅས་
ཡོངས་
ལས
ཙམ་
གྱིས་
དེ་
ཡང་
མཐའ་དག་
ཏུ་
ཉིད་
ས
ཏེ་
གྱི་
སྤྱི
དེ
ཀ་
ཡིན་
ཞིང་
འདི་
རུང་
རང་
ཞིག་
སྟེ
སྟེ་
ན་རེ
ངམ
ཤིང་
དག་
ཏོ
རེ་
འང་
ཀྱང་
ལགས་པ
ཚུ
དོ
ཡིན་པ
རེ
ན་རེ་
ཨེ་
ཚང་མ
ཐམས་ཅད་
དམ་
འོ་
ཅིག་
གྱིན་
ཡིན
ན
ཁོ་ན་
འམ་
ཀྱིན་
ལོ
ཀྱིས
བས་
ལགས་
ཤིག
གིས
ཀི་
སྣ་ཚོགས་
རྣམས
སྙེད་པ
ཡིས་
གྱི
གི
བམ་
ཤིག་
རེ་རེ་
ནམ
མིན་
ནམ་
ངམ་
རུ་
འགའ་
ཀུན
ཤས་
ཏུ
ཡིས
གིན་
གམ་
འོ
ཡིན་པ་
མིན
ལགས
གྱིས
ཅང་
འགའ
སམ་
ཞིག
འང
ལས་ཆེ་
འཕྲལ་
བར་
རུ
དང
ཡ
འག
སམ
ཀ
ཅུང་ཟད་
ཅིག
ཉིད
དུ་མ
མ
ཡིན་བ
འམ
མམ
དམ
དག
ཁོ་ན
ཀྱི
ལམ
ཕྱི་
ནང་
ཙམ
ནོ་
སོ་
རམ་
བོ་
ཨང་
ཕྱི
ཏོ་
ཚོ
ལ་ལ་
ཚོ་
ཅིང
མ་གི་
གེ
གོ
ཡིན་ལུགས་
རོ་
བོ
ལགས་པ་
པས
རབ་
འི
རམ
བས
གཞན
སྙེད་པ་
འབའ་
མཾ་
པོ
ག་
ག
གམ
སྤྱི་
བམ
མོ་
ཙམ་པ་
ཤ་སྟག་
མམ་
རེ་རེ
སྙེད
ཏམ་
ངོ
གྲང་
ཏ་རེ
ཏམ
ཁ་
ངེ་
ཅོག་
རིལ་
ཉུང་ཤས་
གིང་
ཚ་
ཀྱང
""".split()
)



from ...attrs import LIKE_NUM

# reference 1: https://en.wikipedia.org/wiki/Tibetan_numerals

_num_words = [
    "ཀླད་ཀོར་",
    "གཅིག་",
    "གཉིས་",
    "གསུམ་",
    "བཞི་",
    "ལྔ་",
    "དྲུག་",
    "བདུན་",
    "བརྒྱད་",
    "དགུ་",
    "བཅུ་",
    "བཅུ་གཅིག་",
    "བཅུ་གཉིས་",
    "བཅུ་གསུམ་",
    "བཅུ་བཞི་",
    "བཅུ་ལྔ་",
    "བཅུ་དྲུག་",
    "བཅུ་བདུན་",
    "བཅུ་པརྒྱད",
    "བཅུ་དགུ་",
    "ཉི་ཤུ་",
    "སུམ་ཅུ",
    "བཞི་བཅུ",
    "ལྔ་བཅུ",
    "དྲུག་ཅུ",
    "བདུན་ཅུ",
    "བརྒྱད་ཅུ",
    "དགུ་བཅུ",
    "བརྒྱ་",
    "སྟོང་",
    "ཁྲི་",
    "ས་ཡ་",
    "	བྱེ་བ་",
    "དུང་ཕྱུར་",
    "ཐེར་འབུམ་",
    "ཐེར་འབུམ་ཆེན་པོ་",
    "ཁྲག་ཁྲིག་",
    "ཁྲག་ཁྲིག་ཆེན་པོ་",
]


def like_num(text):
    """
    Check if text resembles a number
    """
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text in _num_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



"""Centralized registry population for spaCy config

This module centralizes registry decorations to prevent circular import issues
with Cython annotation changes from __future__ import annotations. Functions
remain in their original locations, but decoration is moved here.

Component definitions and registrations are in spacy/pipeline/factories.py
"""

# Global flag to track if registry has been populated
REGISTRY_POPULATED = False


def populate_registry() -> None:
    """Populate the registry with all necessary components.

    This function should be called before accessing the registry, to ensure
    it's populated. The function uses a global flag to prevent repopulation.
    """
    global REGISTRY_POPULATED
    if REGISTRY_POPULATED:
        return

    # Import all necessary modules
    from .lang.ja import create_tokenizer as create_japanese_tokenizer
    from .lang.ko import create_tokenizer as create_korean_tokenizer
    from .lang.th import create_thai_tokenizer
    from .lang.vi import create_vietnamese_tokenizer
    from .lang.zh import create_chinese_tokenizer
    from .language import load_lookups_data
    from .matcher.levenshtein import make_levenshtein_compare
    from .ml.models.entity_linker import (
        create_candidates,
        create_candidates_batch,
        empty_kb,
        empty_kb_for_config,
        load_kb,
    )
    from .pipeline.attributeruler import make_attribute_ruler_scorer
    from .pipeline.dep_parser import make_parser_scorer

    # Import the functions we refactored by removing direct registry decorators
    from .pipeline.entity_linker import make_entity_linker_scorer
    from .pipeline.entityruler import (
        make_entity_ruler_scorer as make_entityruler_scorer,
    )
    from .pipeline.lemmatizer import make_lemmatizer_scorer
    from .pipeline.morphologizer import make_morphologizer_scorer
    from .pipeline.ner import make_ner_scorer
    from .pipeline.senter import make_senter_scorer
    from .pipeline.span_finder import make_span_finder_scorer
    from .pipeline.span_ruler import (
        make_overlapping_labeled_spans_scorer,
        make_preserve_existing_ents_filter,
        make_prioritize_new_ents_filter,
    )
    from .pipeline.spancat import (
        build_ngram_range_suggester,
        build_ngram_suggester,
        build_preset_spans_suggester,
        make_spancat_scorer,
    )

    # Import all pipeline components that were using registry decorators
    from .pipeline.tagger import make_tagger_scorer
    from .pipeline.textcat import make_textcat_scorer
    from .pipeline.textcat_multilabel import make_textcat_multilabel_scorer
    from .util import make_first_longest_spans_filter, registry

    # Register miscellaneous components
    registry.misc("spacy.first_longest_spans_filter.v1")(
        make_first_longest_spans_filter
    )
    registry.misc("spacy.ngram_suggester.v1")(build_ngram_suggester)
    registry.misc("spacy.ngram_range_suggester.v1")(build_ngram_range_suggester)
    registry.misc("spacy.preset_spans_suggester.v1")(build_preset_spans_suggester)
    registry.misc("spacy.prioritize_new_ents_filter.v1")(
        make_prioritize_new_ents_filter
    )
    registry.misc("spacy.prioritize_existing_ents_filter.v1")(
        make_preserve_existing_ents_filter
    )
    registry.misc("spacy.levenshtein_compare.v1")(make_levenshtein_compare)
    # KB-related registrations
    registry.misc("spacy.KBFromFile.v1")(load_kb)
    registry.misc("spacy.EmptyKB.v2")(empty_kb_for_config)
    registry.misc("spacy.EmptyKB.v1")(empty_kb)
    registry.misc("spacy.CandidateGenerator.v1")(create_candidates)
    registry.misc("spacy.CandidateBatchGenerator.v1")(create_candidates_batch)
    registry.misc("spacy.LookupsDataLoader.v1")(load_lookups_data)

    # Need to get references to the existing functions in registry by importing the function that is there
    # For the registry that was previously decorated

    # Import ML components that use registry
    from .language import create_tokenizer
    from .ml._precomputable_affine import PrecomputableAffine
    from .ml.callbacks import (
        create_models_and_pipes_with_nvtx_range,
        create_models_with_nvtx_range,
    )
    from .ml.extract_ngrams import extract_ngrams
    from .ml.extract_spans import extract_spans

    # Import decorator-removed ML components
    from .ml.featureextractor import FeatureExtractor
    from .ml.models.entity_linker import build_nel_encoder
    from .ml.models.multi_task import (
        create_pretrain_characters,
        create_pretrain_vectors,
    )
    from .ml.models.parser import build_tb_parser_model
    from .ml.models.span_finder import build_finder_model
    from .ml.models.spancat import (
        build_linear_logistic,
        build_mean_max_reducer,
        build_spancat_model,
    )
    from .ml.models.tagger import build_tagger_model
    from .ml.models.textcat import (
        build_bow_text_classifier,
        build_bow_text_classifier_v3,
        build_reduce_text_classifier,
        build_simple_cnn_text_classifier,
        build_text_classifier_lowdata,
        build_text_classifier_v2,
        build_textcat_parametric_attention_v1,
    )
    from .ml.models.tok2vec import (
        BiLSTMEncoder,
        CharacterEmbed,
        MaxoutWindowEncoder,
        MishWindowEncoder,
        MultiHashEmbed,
        build_hash_embed_cnn_tok2vec,
        build_Tok2Vec_model,
        tok2vec_listener_v1,
    )
    from .ml.staticvectors import StaticVectors
    from .ml.tb_framework import TransitionModel
    from .training.augment import (
        create_combined_augmenter,
        create_lower_casing_augmenter,
        create_orth_variants_augmenter,
    )
    from .training.batchers import (
        configure_minibatch,
        configure_minibatch_by_padded_size,
        configure_minibatch_by_words,
    )
    from .training.callbacks import create_copy_from_base_model
    from .training.loggers import console_logger, console_logger_v3

    # Register scorers
    registry.scorers("spacy.tagger_scorer.v1")(make_tagger_scorer)
    registry.scorers("spacy.ner_scorer.v1")(make_ner_scorer)
    # span_ruler_scorer removed as it's not in span_ruler.py
    registry.scorers("spacy.entity_ruler_scorer.v1")(make_entityruler_scorer)
    registry.scorers("spacy.senter_scorer.v1")(make_senter_scorer)
    registry.scorers("spacy.textcat_scorer.v1")(make_textcat_scorer)
    registry.scorers("spacy.textcat_scorer.v2")(make_textcat_scorer)
    registry.scorers("spacy.textcat_multilabel_scorer.v1")(
        make_textcat_multilabel_scorer
    )
    registry.scorers("spacy.textcat_multilabel_scorer.v2")(
        make_textcat_multilabel_scorer
    )
    registry.scorers("spacy.lemmatizer_scorer.v1")(make_lemmatizer_scorer)
    registry.scorers("spacy.span_finder_scorer.v1")(make_span_finder_scorer)
    registry.scorers("spacy.spancat_scorer.v1")(make_spancat_scorer)
    registry.scorers("spacy.entity_linker_scorer.v1")(make_entity_linker_scorer)
    registry.scorers("spacy.overlapping_labeled_spans_scorer.v1")(
        make_overlapping_labeled_spans_scorer
    )
    registry.scorers("spacy.attribute_ruler_scorer.v1")(make_attribute_ruler_scorer)
    registry.scorers("spacy.parser_scorer.v1")(make_parser_scorer)
    registry.scorers("spacy.morphologizer_scorer.v1")(make_morphologizer_scorer)

    # Register tokenizers
    registry.tokenizers("spacy.Tokenizer.v1")(create_tokenizer)
    registry.tokenizers("spacy.ja.JapaneseTokenizer")(create_japanese_tokenizer)
    registry.tokenizers("spacy.zh.ChineseTokenizer")(create_chinese_tokenizer)
    registry.tokenizers("spacy.ko.KoreanTokenizer")(create_korean_tokenizer)
    registry.tokenizers("spacy.vi.VietnameseTokenizer")(create_vietnamese_tokenizer)
    registry.tokenizers("spacy.th.ThaiTokenizer")(create_thai_tokenizer)

    # Register tok2vec architectures we've modified
    registry.architectures("spacy.Tok2VecListener.v1")(tok2vec_listener_v1)
    registry.architectures("spacy.HashEmbedCNN.v2")(build_hash_embed_cnn_tok2vec)
    registry.architectures("spacy.Tok2Vec.v2")(build_Tok2Vec_model)
    registry.architectures("spacy.MultiHashEmbed.v2")(MultiHashEmbed)
    registry.architectures("spacy.CharacterEmbed.v2")(CharacterEmbed)
    registry.architectures("spacy.MaxoutWindowEncoder.v2")(MaxoutWindowEncoder)
    registry.architectures("spacy.MishWindowEncoder.v2")(MishWindowEncoder)
    registry.architectures("spacy.TorchBiLSTMEncoder.v1")(BiLSTMEncoder)
    registry.architectures("spacy.EntityLinker.v2")(build_nel_encoder)
    registry.architectures("spacy.TextCatCNN.v2")(build_simple_cnn_text_classifier)
    registry.architectures("spacy.TextCatBOW.v2")(build_bow_text_classifier)
    registry.architectures("spacy.TextCatBOW.v3")(build_bow_text_classifier_v3)
    registry.architectures("spacy.TextCatEnsemble.v2")(build_text_classifier_v2)
    registry.architectures("spacy.TextCatLowData.v1")(build_text_classifier_lowdata)
    registry.architectures("spacy.TextCatParametricAttention.v1")(
        build_textcat_parametric_attention_v1
    )
    registry.architectures("spacy.TextCatReduce.v1")(build_reduce_text_classifier)
    registry.architectures("spacy.SpanCategorizer.v1")(build_spancat_model)
    registry.architectures("spacy.SpanFinder.v1")(build_finder_model)
    registry.architectures("spacy.TransitionBasedParser.v2")(build_tb_parser_model)
    registry.architectures("spacy.PretrainVectors.v1")(create_pretrain_vectors)
    registry.architectures("spacy.PretrainCharacters.v1")(create_pretrain_characters)
    registry.architectures("spacy.Tagger.v2")(build_tagger_model)

    # Register layers
    registry.layers("spacy.FeatureExtractor.v1")(FeatureExtractor)
    registry.layers("spacy.extract_spans.v1")(extract_spans)
    registry.layers("spacy.extract_ngrams.v1")(extract_ngrams)
    registry.layers("spacy.LinearLogistic.v1")(build_linear_logistic)
    registry.layers("spacy.mean_max_reducer.v1")(build_mean_max_reducer)
    registry.layers("spacy.StaticVectors.v2")(StaticVectors)
    registry.layers("spacy.PrecomputableAffine.v1")(PrecomputableAffine)
    registry.layers("spacy.CharEmbed.v1")(CharacterEmbed)
    registry.layers("spacy.TransitionModel.v1")(TransitionModel)

    # Register callbacks
    registry.callbacks("spacy.copy_from_base_model.v1")(create_copy_from_base_model)
    registry.callbacks("spacy.models_with_nvtx_range.v1")(create_models_with_nvtx_range)
    registry.callbacks("spacy.models_and_pipes_with_nvtx_range.v1")(
        create_models_and_pipes_with_nvtx_range
    )

    # Register loggers
    registry.loggers("spacy.ConsoleLogger.v2")(console_logger)
    registry.loggers("spacy.ConsoleLogger.v3")(console_logger_v3)

    # Register batchers
    registry.batchers("spacy.batch_by_padded.v1")(configure_minibatch_by_padded_size)
    registry.batchers("spacy.batch_by_words.v1")(configure_minibatch_by_words)
    registry.batchers("spacy.batch_by_sequence.v1")(configure_minibatch)

    # Register augmenters
    registry.augmenters("spacy.combined_augmenter.v1")(create_combined_augmenter)
    registry.augmenters("spacy.lower_case.v1")(create_lower_casing_augmenter)
    registry.augmenters("spacy.orth_variants.v1")(create_orth_variants_augmenter)

    # Set the flag to indicate that the registry has been populated
    REGISTRY_POPULATED = True
